from textwrap import dedent

import pytest

from hpcflow.app import app as hf
from hpcflow.sdk.config.errors import UnknownMetaTaskConstitutiveSchema


def test_basic_meta_task_workflow(new_null_config, tmp_path):
    wk_yaml = dedent(
        """\
        name: test_meta_task
        template_components:
          task_schemas:
            - objective: s0
              inputs: 
                - parameter: p1
              outputs:
                - parameter: p2
              actions:
                - commands:
                  - command: echo "$((<<parameter:p1>> + 1))"
                    stdout: <<int(parameter:p2)>>
  
            - objective: s1
              inputs: 
                - parameter: p2
                - parameter: p2b
              outputs:
                - parameter: p3
              actions:
                - commands:
                  - command: echo "$((<<parameter:p2>> + <<parameter:p2b>>))"
                    stdout: <<int(parameter:p3)>>
  
            - objective: s2
              inputs: 
                - parameter: p3
              outputs:
                - parameter: p4
              actions:
                - commands:
                  - command: echo "$((<<parameter:p3>> + 1))"
                    stdout: <<int(parameter:p4)>>
  
            - objective: s3
              inputs: 
                - parameter: p4
              outputs:
                - parameter: p5
              actions:
                - commands:
                  - command: echo "$((<<parameter:p4>> + 1))"
                    stdout: <<int(parameter:p5)>>

          meta_task_schemas:
            - objective: system_analysis
              inputs:
                - parameter: p2
              outputs:
                - parameter: p4

        meta_tasks:
          system_analysis:
            - schema: s1
              inputs:
                p2b: 220
            - schema: s2

        tasks:
          - schema: s0
            inputs:
              p1: 100
          - schema: system_analysis
          - schema: s3
        """
    )
    wk = hf.Workflow.from_YAML_string(wk_yaml, path=tmp_path)

    # basic check of param dependendices
    s0_di = wk.tasks.s0.elements[0].get_data_idx()
    s1_di = wk.tasks.s1.elements[0].get_data_idx()
    s2_di = wk.tasks.s2.elements[0].get_data_idx()
    s3_di = wk.tasks.s3.elements[0].get_data_idx()

    assert s0_di["outputs.p2"] == s1_di["inputs.p2"]
    assert s1_di["outputs.p3"] == s2_di["inputs.p3"]
    assert s2_di["outputs.p4"] == s3_di["inputs.p4"]


def test_basic_meta_task_workflow_API(new_null_config, tmp_path):
    """as above but using Python API."""
    # normal task schemas:
    s0 = hf.TaskSchema(
        objective="s0",
        inputs=[hf.SchemaInput("p1")],
        outputs=[hf.SchemaOutput("p2")],
        actions=[
            hf.Action(
                commands=[
                    hf.Command(
                        command='echo "$((<<parameter:p1>> + 1))"',
                        stdout="<<int(parameter:p2)>>",
                    )
                ]
            )
        ],
    )
    s1 = hf.TaskSchema(
        objective="s1",
        inputs=[hf.SchemaInput("p2")],
        outputs=[hf.SchemaOutput("p3")],
        actions=[
            hf.Action(
                commands=[
                    hf.Command(
                        command='echo "$((<<parameter:p2>> + 1))"',
                        stdout="<<int(parameter:p3)>>",
                    )
                ]
            )
        ],
    )
    s2 = hf.TaskSchema(
        objective="s2",
        inputs=[hf.SchemaInput("p3")],
        outputs=[hf.SchemaOutput("p4")],
        actions=[
            hf.Action(
                commands=[
                    hf.Command(
                        command='echo "$((<<parameter:p3>> + 1))"',
                        stdout="<<int(parameter:p4)>>",
                    )
                ]
            )
        ],
    )
    s3 = hf.TaskSchema(
        objective="s3",
        inputs=[hf.SchemaInput("p4")],
        outputs=[hf.SchemaOutput("p5")],
        actions=[
            hf.Action(
                commands=[
                    hf.Command(
                        command='echo "$((<<parameter:p4>> + 1))"',
                        stdout="<<int(parameter:p5)>>",
                    )
                ]
            )
        ],
    )

    # meta=task schema:
    ms = hf.MetaTaskSchema(
        objective="system_analysis",
        inputs=[hf.SchemaInput("p2")],
        outputs=[hf.SchemaOutput("p4")],
    )

    # meta-task:
    m1 = hf.MetaTask(
        schema=ms,
        tasks=[
            hf.Task(schema=s1),
            hf.Task(schema=s2),
        ],
    )

    # workflow template tasks list:
    tasks = [
        hf.Task(schema=s0, inputs={"p1": 100}),
        m1,
        hf.Task(schema=s3),
    ]

    wk = hf.Workflow.from_template_data(
        template_name="meta_task_workflow",
        tasks=tasks,
        path=tmp_path,
    )

    # basic check of param dependendices
    s0_di = wk.tasks.s0.elements[0].get_data_idx()
    s1_di = wk.tasks.s1.elements[0].get_data_idx()
    s2_di = wk.tasks.s2.elements[0].get_data_idx()
    s3_di = wk.tasks.s3.elements[0].get_data_idx()

    assert s0_di["outputs.p2"] == s1_di["inputs.p2"]
    assert s1_di["outputs.p3"] == s2_di["inputs.p3"]
    assert s2_di["outputs.p4"] == s3_di["inputs.p4"]


def test_meta_task_custom_parametrisation(new_null_config, tmp_path):
    """test customising the parametrisation of inputs, sequences, and resources within the
    `tasks` list."""
    wk_yaml = dedent(
        """\
        name: test_metatask_multi_element_sets_custom_parametrisation
        template_components:
          task_schemas:
            - objective: s1
              inputs: 
                - parameter: p1
                - parameter: p2
              outputs:
                - parameter: p3
              actions:
                - commands:
                  - command: echo "$((<<parameter:p1>> + <<parameter:p2>>))"
                    stdout: <<int(parameter:p3)>>

          meta_task_schemas:
            - objective: system_analysis
              inputs:
                - parameter: p1
                - parameter: p2
              outputs:
                - parameter: p3

        meta_tasks:
          system_analysis:
            - schema: s1
              element_sets:
                - inputs:
                    p1: 100
                    p2: 200
                - inputs:
                    p1: 100
                  sequences:
                    - path: inputs.p2
                      values: [200, 201]
        tasks:
          - schema: system_analysis
            inputs:
              s1: # should apply to first element set by default
                p1: 101
            resources:
              s1: # should apply to first element set by default
                any:
                  num_cores: 2
          - schema: system_analysis
            inputs:
              s1.0: # applies to first element set of s1
                p1: 102
              s1.1: # applies to second element set of s1
                p1: 103
            sequences:
              s1.1: # sequences list in second element set is replaced with this list:
                - path: inputs.p2
                  values: [300, 301]
        """
    )
    wk = hf.Workflow.from_YAML_string(wk_yaml, path=tmp_path)

    assert wk.tasks.s1_1.template.element_sets[0].resources[0].num_cores == 2  # modified
    assert (
        wk.tasks.s1_2.template.element_sets[0].resources[0].num_cores is None
    )  # unaffected

    assert wk.tasks.s1_1.template.element_sets[0].inputs[0].value == 101  # modified
    assert wk.tasks.s1_1.template.element_sets[0].inputs[1].value == 200  # unaffected
    assert wk.tasks.s1_1.template.element_sets[1].sequences[0].values == [
        200,
        201,
    ]  # unaffected

    assert wk.tasks.s1_2.template.element_sets[0].inputs[0].value == 102  # modified
    assert wk.tasks.s1_2.template.element_sets[1].inputs[0].value == 103  # modified
    assert wk.tasks.s1_2.template.element_sets[1].sequences[0].values == [
        300,
        301,
    ]  # modified


def test_meta_task_custom_parametrisation_raises_on_bad_schema_name(
    new_null_config, tmp_path
):
    wk_yaml = dedent(
        """\
      name: test_metatask_raise_on_bad_schema_name
      template_components:
        task_schemas:
          - objective: s1
            inputs: 
              - parameter: p1
              - parameter: p2
            outputs:
              - parameter: p3
            actions:
              - commands:
                - command: echo "$((<<parameter:p1>> + <<parameter:p2>>))"
                  stdout: <<int(parameter:p3)>>

        meta_task_schemas:
          - objective: system_analysis
            inputs:
              - parameter: p1
              - parameter: p2
            outputs:
              - parameter: p3

      meta_tasks:
        system_analysis:
          - schema: s1
            element_sets:
              - inputs:
                  p1: 100
                  p2: 200
              - inputs:
                  p1: 100
                sequences:
                  - path: inputs.p2
                    values: [200, 201]
      tasks:
        - schema: system_analysis
          resources:
            BAD_SCHEMA_NAME: # should raise!
              any:
                num_cores: 2
      """
    )
    with pytest.raises(UnknownMetaTaskConstitutiveSchema):
        wk = hf.Workflow.from_YAML_string(wk_yaml, path=tmp_path)



"""
General model of a searchable serializable list.
"""

from __future__ import annotations
from collections import defaultdict
from collections.abc import Mapping, Sequence
import copy
import sys
from types import SimpleNamespace
from typing import Generic, TypeVar, cast, overload, TYPE_CHECKING
from typing_extensions import override

from hpcflow.sdk.core.json_like import ChildObjectSpec, JSONLike

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator
    from typing import Any, ClassVar, Literal
    from typing_extensions import Self, TypeIs
    from zarr import Group  # type: ignore
    from .actions import ActionScope
    from .command_files import FileSpec
    from .environment import Environment, Executable
    from .loop import WorkflowLoop
    from .json_like import JSONable, JSONed
    from .parameters import Parameter, ResourceSpec
    from .task import Task, TaskTemplate, TaskSchema, WorkflowTask, ElementSet
    from .types import Resources
    from .workflow import WorkflowTemplate

T = TypeVar("T")


class ObjectListMultipleMatchError(ValueError):
    """
    Thrown when an object looked up by unique attribute ends up with multiple objects
    being matched.
    """


class ObjectList(JSONLike, Generic[T]):
    """
    A list-like class that provides item access via a `get` method according to
    attributes or dict-keys.

    Parameters
    ----------
    objects : sequence
        List of values of some type.
    descriptor : str
        Descriptive name for objects in the list.
    """

    # This would be in the docstring except it renders really wrongly!
    # Type Parameters
    # ---------------
    # T
    #     The type of elements of the list.

    def __init__(self, objects: Iterable[T], descriptor: str | None = None):
        self._objects = list(objects)
        self._descriptor = descriptor or "object"
        self._object_is_dict: bool = False
        self._validate()

    def __deepcopy__(self, memo: dict[int, Any]) -> Self:
        obj = self.__class__(copy.deepcopy(self._objects, memo))
        obj._descriptor = self._descriptor
        obj._object_is_dict = self._object_is_dict
        return obj

    def _validate(self):
        for idx, obj in enumerate(self._objects):
            if isinstance(obj, dict):
                obj = SimpleNamespace(**obj)
                self._object_is_dict = True
                self._objects[idx] = obj

    def __len__(self):
        return len(self._objects)

    def __repr__(self):
        return repr(self._objects)

    def __str__(self):
        return str([self._get_item(obj) for obj in self._objects])

    def __iter__(self) -> Iterator[T]:
        if self._object_is_dict:
            return iter(self._get_item(obj) for obj in self._objects)
        else:
            return self._objects.__iter__()

    @overload
    def __getitem__(self, key: int) -> T: ...

    @overload
    def __getitem__(self, key: slice) -> list[T]: ...

    def __getitem__(self, key: int | slice) -> T | list[T]:
        """Provide list-like index access."""
        if isinstance(key, slice):
            return list(map(self._get_item, self._objects.__getitem__(key)))
        else:
            return self._get_item(self._objects.__getitem__(key))

    def __contains__(self, item: T) -> bool:
        if self._objects:
            if type(item) is type(self._get_item(self._objects[0])):
                return self._objects.__contains__(item)
        return False

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, self.__class__) and self._objects == other._objects

    def _get_item(self, obj: T):
        if self._object_is_dict:
            return obj.__dict__
        else:
            return obj

    def _get_obj_attr(self, obj: T, attr: str):
        """Overriding this function allows control over how the `get` functions behave."""
        return getattr(obj, attr)

    def __specified_objs(self, objs: Iterable[T], kwargs: dict[str, Any]) -> Iterator[T]:
        for obj in objs:
            for k, v in kwargs.items():
                try:
                    if self._get_obj_attr(obj, k) != v:
                        break
                except (AttributeError, KeyError):
                    break
            else:
                yield obj

    def _get_all_from_objs(self, objs: Iterable[T], **kwargs):
        # narrow down according to kwargs:
        return [self._get_item(obj) for obj in self.__specified_objs(objs, kwargs)]

    def get_all(self, **kwargs):
        """Get one or more objects from the object list, by specifying the value of the
        access attribute, and optionally additional keyword-argument attribute values."""

        return self._get_all_from_objs(self._objects, **kwargs)

    def _validate_get(self, result: Sequence[T], kwargs: dict[str, Any]):
        if not result:
            available: list[dict[str, Any]] = []
            for obj in self._objects:
                attr_vals: dict[str, Any] = {}
                for k in kwargs:
                    try:
                        attr_vals[k] = self._get_obj_attr(obj, k)
                    except (AttributeError, KeyError):
                        continue
                available.append(attr_vals)
            raise ValueError(
                f"No {self._descriptor} objects with attributes: {kwargs}. Available "
                f"objects have attributes: {tuple(available)!r}."
            )

        elif len(result) > 1:
            raise ObjectListMultipleMatchError(
                f"Multiple objects with attributes: {kwargs}."
            )

        return result[0]

    def get(self, **kwargs):
        """Get a single object from the object list, by specifying the value of the access
        attribute, and optionally additional keyword-argument attribute values."""
        return self._validate_get(self.get_all(**kwargs), kwargs)

    @overload
    def add_object(
        self, obj: T, index: int = -1, *, skip_duplicates: Literal[False] = False
    ) -> int: ...

    @overload
    def add_object(
        self, obj: T, index: int = -1, *, skip_duplicates: Literal[True]
    ) -> int | None: ...

    def add_object(
        self, obj: T, index: int = -1, *, skip_duplicates: bool = False
    ) -> None | int:
        """
        Add an object to this object list.

        Parameters
        ----------
        obj:
            The object to add.
        index:
            Where to add it. Omit to append.
        skip_duplicates:
            If true, don't add the object if it is already in the list.

        Returns
        -------
            The index of the added object, or ``None`` if the object was not added.
        """
        if skip_duplicates and obj in self:
            return None

        if index < 0:
            index += len(self) + 1

        if self._object_is_dict:
            obj = cast("T", SimpleNamespace(**cast("dict", obj)))

        self._objects = self._objects[:index] + [obj] + self._objects[index:]
        self._validate()
        return index


class DotAccessAttributeError(AttributeError):
    def __init__(self, name: str, obj: DotAccessObjectList) -> None:
        msg = f"{obj._descriptor.title()} {name!r} does not exist. "
        if obj._objects:
            attr = obj._access_attribute
            obj_list = (f'"{getattr(obj, attr)}"' for obj in obj._objects)
            msg += f"Available {obj._descriptor}s are: {', '.join(obj_list)}."
        else:
            msg += "The object list is empty."
        if sys.version_info >= (3, 10):
            super().__init__(msg, name=name, obj=obj)
        else:
            super().__init__(msg)


class DotAccessObjectList(ObjectList[T], Generic[T]):
    """
    Provide dot-notation access via an access attribute for the case where the access
    attribute uniquely identifies a single object.

    Parameters
    ----------
    _objects:
        The objects in the list.
    access_attribute:
        The main attribute for selection and filtering. A unique property.
    descriptor: str
        Descriptive name for the objects in the list.
    """

    # This would be in the docstring except it renders really wrongly!
    # Type Parameters
    # ---------------
    # T
    #     The type of elements of the list.

    # access attributes must not be named after any "public" methods, to avoid confusion!
    _pub_methods: ClassVar[tuple[str, ...]] = (
        "get",
        "get_all",
        "add_object",
        "add_objects",
    )

    def __init__(
        self, _objects: Iterable[T], access_attribute: str, descriptor: str | None = None
    ):
        self._access_attribute = access_attribute
        self._index: Mapping[str, Sequence[int]]
        super().__init__(_objects, descriptor=descriptor)
        self._update_index()

    def __deepcopy__(self, memo: dict[int, Any]) -> Self:
        obj = self.__class__(copy.deepcopy(self._objects, memo), self._access_attribute)
        obj._descriptor = self._descriptor
        obj._object_is_dict = self._object_is_dict
        return obj

    def _validate(self) -> None:
        for idx, obj in enumerate(self._objects):
            if not hasattr(obj, self._access_attribute):
                raise TypeError(
                    f"Object {idx} does not have attribute {self._access_attribute!r}."
                )
            value = getattr(obj, self._access_attribute)
            if value in self._pub_methods:
                raise ValueError(
                    f"Access attribute {self._access_attribute!r} for object index {idx} "
                    f"cannot be the same as any of the methods of "
                    f"{self.__class__.__name__!r}, which are: {self._pub_methods!r}."
                )
        super()._validate()

    def _update_index(self) -> None:
        """For quick look-up by access attribute."""

        _index: dict[str, list[int]] = defaultdict(list)
        for idx, obj in enumerate(self._objects):
            attr_val: str = getattr(obj, self._access_attribute)
            try:
                _index[attr_val].append(idx)
            except TypeError:
                raise TypeError(
                    f"Access attribute values ({self._access_attribute!r}) must be hashable."
                )
        self._index = _index

    def __getattr__(self, attribute: str):
        if idx := self._index.get(attribute):
            if len(idx) > 1:
                raise ValueError(
                    f"Multiple objects with access attribute: {attribute!r}."
                )
            return self._get_item(self._objects[idx[0]])
        elif not attribute.startswith("__"):
            raise DotAccessAttributeError(attribute, self)
        else:
            raise AttributeError

    def __dir__(self) -> Iterator[str]:
        yield from super().__dir__()
        yield from (getattr(obj, self._access_attribute) for obj in self._objects)

    def list_attrs(self) -> tuple[str, ...]:
        """Get a tuple of the unique access-attribute values of the constituent objects."""
        return tuple(self._index)

    def get(self, access_attribute_value: str | None = None, **kwargs) -> T:
        """
        Get an object from this list that matches the given criteria.
        """
        vld_get_kwargs = kwargs
        if access_attribute_value is not None:
            vld_get_kwargs = {self._access_attribute: access_attribute_value, **kwargs}

        return self._validate_get(
            self.get_all(access_attribute_value=access_attribute_value, **kwargs),
            vld_get_kwargs,
        )

    def get_all(self, access_attribute_value: str | None = None, **kwargs):
        """
        Get all objects in this list that match the given criteria.
        """
        # use the index to narrow down the search first:
        if access_attribute_value is not None:
            if (all_idx := self._index.get(access_attribute_value)) is None:
                raise ValueError(
                    f"Value {access_attribute_value!r} does not match the value of any "
                    f"object's attribute {self._access_attribute!r}. Available attribute "
                    f"values are: {self.list_attrs()!r}."
                )
            all_objs: Iterable[T] = (self._objects[idx] for idx in all_idx)
        else:
            all_objs = self._objects

        return self._get_all_from_objs(all_objs, **kwargs)

    @overload
    def add_object(
        self, obj: T, index: int = -1, *, skip_duplicates: Literal[False] = False
    ) -> int: ...

    @overload
    def add_object(
        self, obj: T, index: int = -1, *, skip_duplicates: Literal[True]
    ) -> int | None: ...

    def add_object(
        self, obj: T, index: int = -1, *, skip_duplicates: bool = False
    ) -> int | None:
        """
        Add an object to this list.
        """
        if skip_duplicates:
            new_index = super().add_object(obj, index, skip_duplicates=True)
        else:
            new_index = super().add_object(obj, index)
        self._update_index()
        return new_index

    def add_objects(
        self, objs: Iterable[T], index: int = -1, *, skip_duplicates: bool = False
    ) -> int:
        """
        Add multiple objects to the list.
        """
        if skip_duplicates:
            for obj in objs:
                if (i := self.add_object(obj, index, skip_duplicates=True)) is not None:
                    index = i + 1
        else:
            for obj in objs:
                index = self.add_object(obj, index) + 1
        return index


class AppDataList(DotAccessObjectList[T], Generic[T]):
    """
    An application-aware object list.

    Type Parameters
    ---------------
    T
        The type of elements of the list.
    """

    @override
    def _postprocess_to_dict(self, d: dict[str, Any]) -> dict[str, Any]:
        d = super()._postprocess_to_dict(d)
        return {"_objects": d["_objects"]}

    @classmethod
    def _get_default_shared_data(cls) -> Mapping[str, ObjectList[JSONable]]:
        return cls._app._shared_data

    @overload
    @classmethod
    def from_json_like(
        cls,
        json_like: str,
        shared_data: Mapping[str, ObjectList[JSONable]] | None = None,
        is_hashed: bool = False,
    ) -> Self | None: ...

    @overload
    @classmethod
    def from_json_like(
        cls,
        json_like: Mapping[str, JSONed] | Sequence[Mapping[str, JSONed]],
        shared_data: Mapping[str, ObjectList[JSONable]] | None = None,
        is_hashed: bool = False,
    ) -> Self: ...

    @overload
    @classmethod
    def from_json_like(
        cls,
        json_like: None,
        shared_data: Mapping[str, ObjectList[JSONable]] | None = None,
        is_hashed: bool = False,
    ) -> None: ...

    @classmethod
    def from_json_like(
        cls,
        json_like: str | Mapping[str, JSONed] | Sequence[Mapping[str, JSONed]] | None,
        shared_data: Mapping[str, ObjectList[JSONable]] | None = None,
        is_hashed: bool = False,
    ) -> Self | None:
        """
        Make an instance of this class from JSON (or YAML) data.

        Parameters
        ----------
        json_like:
            The data to deserialise.
        shared_data:
            Shared context data.
        is_hashed:
            If True, accept a dict whose keys are hashes of the dict values.

        Returns
        -------
            The deserialised object.
        """
        if is_hashed:
            assert isinstance(json_like, Mapping)
            return super().from_json_like(
                [
                    {**cast("Mapping", obj_js), "_hash_value": hash_val}
                    for hash_val, obj_js in json_like.items()
                ],
                shared_data=shared_data,
            )
        else:
            return super().from_json_like(json_like, shared_data=shared_data)

    def _remove_object(self, index: int):
        self._objects.pop(index)
        self._update_index()


class TaskList(AppDataList["Task"]):
    """A list-like container for a task-like list with dot-notation access by task
    unique-name.

    Parameters
    ----------
    _objects: list[~hpcflow.app.Task]
        The tasks in this list.
    """

    _child_objects: ClassVar[tuple[ChildObjectSpec, ...]] = (
        ChildObjectSpec(
            name="_objects",
            class_name="Task",
            is_multiple=True,
            is_single_attribute=True,
        ),
    )

    def __init__(self, _objects: Iterable[Task]):
        super().__init__(_objects, access_attribute="unique_name", descriptor="task")


class TaskTemplateList(AppDataList["TaskTemplate"]):
    """A list-like container for a task-like list with dot-notation access by task
    unique-name.

    Parameters
    ----------
    _objects: list[~hpcflow.app.TaskTemplate]
        The task templates in this list.
    """

    _child_objects: ClassVar[tuple[ChildObjectSpec, ...]] = (
        ChildObjectSpec(
            name="_objects",
            class_name="TaskTemplate",
            is_multiple=True,
            is_single_attribute=True,
        ),
    )

    def __init__(self, _objects: Iterable[TaskTemplate]):
        super().__init__(_objects, access_attribute="name", descriptor="task template")


class TaskSchemasList(AppDataList["TaskSchema"]):
    """A list-like container for a task schema list with dot-notation access by task
    schema unique-name.

    Parameters
    ----------
    _objects: list[~hpcflow.app.TaskSchema]
        The task schemas in this list.
    """

    _child_objects: ClassVar[tuple[ChildObjectSpec, ...]] = (
        ChildObjectSpec(
            name="_objects",
            class_name="TaskSchema",
            is_multiple=True,
            is_single_attribute=True,
        ),
    )

    def __init__(self, _objects: Iterable[TaskSchema]):
        super().__init__(_objects, access_attribute="name", descriptor="task schema")


class GroupList(AppDataList["Group"]):
    """A list-like container for the task schema group list with dot-notation access by
    group name.

    Parameters
    ----------
    _objects: list[Group]
        The groups in this list.
    """

    _child_objects: ClassVar[tuple[ChildObjectSpec, ...]] = (
        ChildObjectSpec(
            name="_objects",
            class_name="Group",
            is_multiple=True,
            is_single_attribute=True,
        ),
    )

    def __init__(self, _objects: Iterable[Group]):
        super().__init__(_objects, access_attribute="name", descriptor="group")


class EnvironmentsList(AppDataList["Environment"]):
    """
    A list-like container for environments with dot-notation access by name.

    Parameters
    ----------
    _objects: list[~hpcflow.app.Environment]
        The environments in this list.
    """

    _child_objects: ClassVar[tuple[ChildObjectSpec, ...]] = (
        ChildObjectSpec(
            name="_objects",
            class_name="Environment",
            is_multiple=True,
            is_single_attribute=True,
        ),
    )

    def __init__(self, _objects: Iterable[Environment]):
        super().__init__(_objects, access_attribute="name", descriptor="environment")

    def _get_obj_attr(self, obj: Environment, attr: str):
        """Overridden to lookup objects via the `specifiers` dict attribute"""
        if attr in ("name", "_hash_value"):
            return getattr(obj, attr)
        else:
            return obj.specifiers[attr]


class ExecutablesList(AppDataList["Executable"]):
    """
    A list-like container for environment executables with dot-notation access by
    executable label.

    Parameters
    ----------
    _objects: list[~hpcflow.app.Executable]
        The executables in this list.
    """

    #: The environment containing these executables.
    environment: Environment | None = None
    _child_objects: ClassVar[tuple[ChildObjectSpec, ...]] = (
        ChildObjectSpec(
            name="_objects",
            class_name="Executable",
            is_multiple=True,
            is_single_attribute=True,
            parent_ref="_executables_list",
        ),
    )

    def __init__(self, _objects: Iterable[Executable]):
        super().__init__(_objects, access_attribute="label", descriptor="executable")
        self._set_parent_refs()

    def __deepcopy__(self, memo: dict[int, Any]):
        obj = super().__deepcopy__(memo)
        obj.environment = self.environment
        return obj


class ParametersList(AppDataList["Parameter"]):
    """
    A list-like container for parameters with dot-notation access by parameter type.

    Parameters
    ----------
    _objects: list[~hpcflow.app.Parameter]
        The parameters in this list.
    """

    _child_objects: ClassVar[tuple[ChildObjectSpec, ...]] = (
        ChildObjectSpec(
            name="_objects",
            class_name="Parameter",
            is_multiple=True,
            is_single_attribute=True,
        ),
    )

    def __init__(self, _objects: Iterable[Parameter]):
        super().__init__(_objects, access_attribute="typ", descriptor="parameter")

    def __getattr__(self, attribute: str) -> Parameter:
        """Overridden to provide a default Parameter object if none exists."""
        try:
            if not attribute.startswith("__"):
                return super().__getattr__(attribute)
        except (AttributeError, ValueError):
            return self._app.Parameter(typ=attribute)
        raise AttributeError

    def get_all(self, access_attribute_value=None, **kwargs):
        """Overridden to provide a default Parameter object if none exists."""
        typ = access_attribute_value if access_attribute_value else kwargs.get("typ")
        try:
            all_out = super().get_all(access_attribute_value, **kwargs)
        except ValueError:
            return [self._app.Parameter(typ=typ)]
        else:
            # `get_all` will not raise `ValueError` if `access_attribute_value` is
            # None and the parameter `typ` is specified in `kwargs` instead:
            return all_out or [self._app.Parameter(typ=typ)]


class CommandFilesList(AppDataList["FileSpec"]):
    """
    A list-like container for command files with dot-notation access by label.

    Parameters
    ----------
    _objects: list[~hpcflow.app.FileSpec]
        The files in this list.
    """

    _child_objects: ClassVar[tuple[ChildObjectSpec, ...]] = (
        ChildObjectSpec(
            name="_objects",
            class_name="FileSpec",
            is_multiple=True,
            is_single_attribute=True,
        ),
    )

    def __init__(self, _objects: Iterable[FileSpec]):
        super().__init__(_objects, access_attribute="label", descriptor="command file")


class WorkflowTaskList(DotAccessObjectList["WorkflowTask"]):
    """
    A list-like container for workflow tasks with dot-notation access by unique name.

    Parameters
    ----------
    _objects: list[~hpcflow.app.WorkflowTask]
        The tasks in this list.
    """

    def __init__(self, _objects: Iterable[WorkflowTask]):
        super().__init__(_objects, access_attribute="unique_name", descriptor="task")

    def _reindex(self) -> None:
        """Re-assign the WorkflowTask index attributes so they match their order."""
        for idx, item in enumerate(self._objects):
            item._index = idx
        self._update_index()

    def add_object(
        self, obj: WorkflowTask, index: int = -1, skip_duplicates=False
    ) -> int:
        index = super().add_object(obj, index)
        self._reindex()
        return index

    def _remove_object(self, index: int):
        self._objects.pop(index)
        self._reindex()


class WorkflowLoopList(DotAccessObjectList["WorkflowLoop"]):
    """
    A list-like container for workflow loops with dot-notation access by name.

    Parameters
    ----------
    _objects: list[~hpcflow.app.WorkflowLoop]
        The loops in this list.
    """

    def __init__(self, _objects: Iterable[WorkflowLoop]):
        super().__init__(_objects, access_attribute="name", descriptor="loop")

    def _remove_object(self, index: int):
        self._objects.pop(index)


class ResourceList(ObjectList["ResourceSpec"]):
    """
    A list-like container for resources.
    Each contained resource must have a unique scope.

    Parameters
    ----------
    _objects: list[~hpcflow.app.ResourceSpec]
        The resource descriptions in this list.
    """

    _child_objects: ClassVar[tuple[ChildObjectSpec, ...]] = (
        ChildObjectSpec(
            name="_objects",
            class_name="ResourceSpec",
            is_multiple=True,
            is_single_attribute=True,
            dict_key_attr="scope",
            parent_ref="_resource_list",
        ),
    )

    def __init__(self, _objects: Iterable[ResourceSpec]):
        super().__init__(_objects, descriptor="resource specification")
        self._element_set: ElementSet | None = None  # assigned by parent ElementSet
        self._workflow_template: WorkflowTemplate | None = (
            None  # assigned by parent WorkflowTemplate
        )

        # check distinct scopes for each item:
        scopes = [scope.to_string() for scope in self.get_scopes()]
        if len(set(scopes)) < len(scopes):
            raise ValueError(
                "Multiple `ResourceSpec` objects have the same scope. The scopes are "
                f"{scopes!r}."
            )

        self._set_parent_refs()

    def __deepcopy__(self, memo: dict[int, Any]):
        obj = super().__deepcopy__(memo)
        obj._element_set = self._element_set
        obj._workflow_template = self._workflow_template
        return obj

    @property
    def element_set(self) -> ElementSet | None:
        """
        The parent element set, if a child of an element set.
        """
        return self._element_set

    @property
    def workflow_template(self) -> WorkflowTemplate | None:
        """
        The parent workflow template, if a child of a workflow template.
        """
        return self._workflow_template

    def _postprocess_to_json(self, json_like):
        """Convert JSON doc to a dict keyed by action scope (like as can be
        specified in the input YAML) instead of list."""
        return {
            self._app.ActionScope.from_json_like(
                res_spec_js.pop("scope")
            ).to_string(): res_spec_js
            for res_spec_js in json_like
        }

    @staticmethod
    def __ensure_non_persistent(resource_spec: ResourceSpec) -> ResourceSpec:
        """
        For any resources that are persistent, if they have a
        `_resource_list` attribute, this means they are sourced from some
        other persistent workflow, rather than, say, a workflow being
        loaded right now, so make a non-persistent copy

        Part of `normalise`.
        """
        if resource_spec._value_group_idx is not None and (
            resource_spec._resource_list is not None
        ):
            return resource_spec.copy_non_persistent()
        return resource_spec

    @classmethod
    def __is_ResourceSpec(cls, value) -> TypeIs[ResourceSpec]:
        return isinstance(value, cls._app.ResourceSpec)

    @classmethod
    def normalise(cls, resources: Resources) -> Self:
        """Generate from resource-specs specified in potentially several ways."""

        if not resources:
            return cls([cls._app.ResourceSpec()])
        elif isinstance(resources, ResourceList):
            # Already a ResourceList
            return cast("Self", resources)
        elif isinstance(resources, dict):
            return cls.from_json_like(cast("dict", resources))
        elif cls.__is_ResourceSpec(resources):
            return cls([resources])
        else:
            return cls(
                (
                    cls._app.ResourceSpec.from_json_like(cast("dict", res_i))
                    if isinstance(res_i, dict)
                    else cls.__ensure_non_persistent(res_i)
                )
                for res_i in resources
            )

    def get_scopes(self) -> Iterator[ActionScope]:
        """
        Get the scopes of the contained resources.
        """
        for rs in self._objects:
            if rs.scope is not None:
                yield rs.scope

    def __get_for_scope(self, scope: ActionScope):
        try:
            return self.get(scope=scope)
        except ValueError:
            return None

    def __merge(self, our_spec: ResourceSpec | None, other_spec: ResourceSpec):
        """
        Merge two resource specs that have the same scope, or just add the other one to
        the list if we didn't already have it.
        """
        if our_spec is not None:
            for k, v in other_spec._get_members().items():
                if getattr(our_spec, k, None) is None:
                    setattr(our_spec, f"_{k}", copy.deepcopy(v))
        else:
            self.add_object(copy.deepcopy(other_spec))

    def merge_other(self, other: ResourceList):
        """Merge lower-precedence other resource list into this resource list."""
        for scope_i in other.get_scopes():
            self.__merge(self.__get_for_scope(scope_i), other.get(scope=scope_i))

    def merge_one(self, other: ResourceSpec):
        """Merge lower-precedence other resource spec into this resource list.

        This is a simplified version of :py:meth:`merge_other`.
        """
        if other.scope is not None:
            self.__merge(self.__get_for_scope(other.scope), other)


def index(obj_lst: ObjectList[T], obj: T) -> int:
    """
    Get the index of the object in the list.
    The item is checked for by object identity, not equality.
    """
    for idx, item in enumerate(obj_lst._objects):
        if obj is item:
            return idx
    raise ValueError(f"{obj!r} not in list.")



"""Tests concerning the directory structure of a created or submitted workflow"""

import os
from pathlib import Path
import pytest

from hpcflow.sdk.core.test_utils import (
    make_test_data_YAML_workflow,
    make_workflow_to_run_command,
)


@pytest.mark.integration
def test_std_stream_file_not_created(tmp_path, new_null_config):
    """Normally, the app standard stream file should not be written."""
    wk = make_test_data_YAML_workflow("workflow_1.yaml", path=tmp_path)
    wk.submit(wait=True, add_to_known=False)
    run = wk.get_all_EARs()[0]
    std_stream_path = run.get_app_std_path()
    assert not std_stream_path.is_file()


@pytest.mark.integration
def test_std_stream_file_created_on_exception_raised(tmp_path, new_null_config):
    command = 'wkflow_app --std-stream "$HPCFLOW_RUN_STD_PATH" internal noop --raise'
    wk = make_workflow_to_run_command(command=command, path=tmp_path)
    wk.submit(wait=True, add_to_known=False)
    run = wk.get_all_EARs()[0]
    std_stream_path = run.get_app_std_path()
    assert std_stream_path.is_file()
    assert "ValueError: internal noop raised!" in std_stream_path.read_text()



from __future__ import annotations
from pathlib import Path
import pytest
from hpcflow.app import app as hf


@pytest.mark.parametrize("store", ["json", "zarr"])
def test_decode(null_config, tmp_path: Path, store: str):
    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput(parameter=hf.Parameter("p1"))],
        outputs=[hf.SchemaInput(parameter=hf.Parameter("p2"))],
        actions=[
            hf.Action(
                commands=[
                    hf.Command(
                        command="Write-Output (<<parameter:p1>> + 100)",
                        stdout="<<parameter:p2>>",
                    )
                ]
            )
        ],
    )
    wk = hf.Workflow.from_template_data(
        tasks=[hf.Task(schema=s1, inputs=[hf.InputValue("p1", value=101)])],
        loops=[hf.Loop(tasks=[0], num_iterations=1)],
        path=tmp_path,
        template_name="wk0",
        store=store,
    )
    iter_i = wk.tasks[0].elements[0].iterations[0]
    assert iter_i.id_ == 0
    assert iter_i.index == 0
    assert iter_i.EARs_initialised == True
    assert sorted(iter_i.data_idx) == sorted(
        {"inputs.p1": 2, "resources.any": 1, "outputs.p2": 3}
    )
    assert iter_i.loop_idx == {"loop_0": 0}
    assert sorted(iter_i.schema_parameters) == sorted(
        ["resources.any", "inputs.p1", "outputs.p2"]
    )


@pytest.mark.integration
def test_loop_skipped_true_single_action_elements(null_config, tmp_path):
    ts = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput("p1")],
        outputs=[hf.SchemaInput("p1")],
        actions=[
            hf.Action(
                commands=[
                    hf.Command(
                        command="echo $(( <<parameter:p1>> + 100 ))",
                        stdout="<<int(parameter:p1)>>",
                    ),
                ]
            ),
        ],
    )
    loop_term = hf.Rule(path="outputs.p1", condition={"value.equal_to": 300})
    wk = hf.Workflow.from_template_data(
        template_name="test_loop_skipped",
        path=tmp_path,
        tasks=[hf.Task(schema=ts, inputs={"p1": 100})],
        loops=[
            hf.Loop(name="my_loop", tasks=[0], termination=loop_term, num_iterations=3)
        ],
    )
    # loop should terminate after the second iteration; third iteration should
    # be loop-skipped
    wk.submit(wait=True, add_to_known=False, status=False)
    iters = wk.get_all_element_iterations()

    assert not iters[0].loop_skipped
    assert not iters[1].loop_skipped
    assert iters[2].loop_skipped

    # check latest iteration is the latest non-loop-skipped iteration:
    assert wk.tasks[0].elements[0].latest_iteration_non_skipped.id_ == iters[1].id_

    # check element inputs are from latest non-loop-skipped iteration:
    assert wk.tasks[0].elements[0].inputs.p1.value == 200
    assert wk.tasks[0].elements[0].get("inputs.p1") == 200

    # check element outputs are from latest non-loop-skipped iteration:
    assert wk.tasks[0].elements[0].outputs.p1.value == 300
    assert wk.tasks[0].elements[0].get("outputs.p1") == 300



"""
Utilities for working with Zarr.
"""

from __future__ import annotations
from typing import Any
from typing_extensions import Self

import zarr  # type: ignore
import numpy as np

from hpcflow.sdk.core.utils import get_in_container, get_relative_path, set_in_container


#: The basic types that Zarr can handle directly with no special action.
PRIMITIVES: tuple[type, ...] = (
    int,
    float,
    str,
    type(None),
)

#: Maximum nesting depth for encoding.
MAX_DEPTH = 50


def zarr_encode(data, zarr_group: zarr.Group, is_pending_add: bool, is_set: bool):
    """
    Encode data into a zarr group.
    """

    encoded: list[dict] = []

    def encode(obj: Any, path: list) -> Any:
        if len(path) > MAX_DEPTH:
            raise RuntimeError("I'm in too deep!")

        if isinstance(obj, ZarrEncodable):
            return encode(obj.to_dict(), path)
        elif isinstance(obj, (list, tuple, set)):
            out = (encode(item, [*path, idx]) for idx, item in enumerate(obj))
            if isinstance(obj, tuple):
                return tuple(out)
            elif isinstance(obj, set):
                return set(out)
            else:
                return list(out)
        elif isinstance(obj, dict):
            return {
                dct_key: encode(dct_val, [*path, dct_key])
                for dct_key, dct_val in obj.items()
            }
        elif isinstance(obj, PRIMITIVES):
            return obj
        elif isinstance(obj, np.ndarray):
            new_name = str(max((int(i) + 1 for i in zarr_group.keys()), default=0))
            zarr_group.create_dataset(name=new_name, data=obj)
            encoded.append(
                {
                    "path": path,
                    "dataset": new_name,
                }
            )
            return None
        else:
            raise ValueError(f"unserializable type: {type(obj)}")

    zarr_group.attrs["data"] = encode(data, [])
    zarr_group.attrs["encoded"] = encoded
    zarr_group.attrs["is_set"] = is_set
    if is_pending_add:
        zarr_group.attrs["is_pending_add"] = is_pending_add


# TODO: need to separate so this func doesn't need matflow


def _zarr_encode_NEW(
    obj: Any,
    root_group: zarr.Group,
    arr_path: str,
) -> tuple[Any, list[list]]:
    """
    Save arbitrarily-nested Python-primitive, `ZarrEncodable` and numpy array objects into
    Zarr.

    Parameters
    ----------
    obj:
        Object to encode.
    root_group:
        Parent Zarr group into which new Zarr arrays will be added (at `arr_path`).
    arr_path:
        Path relative to `root_group` into which new Zarr arrays will be added.

    Returns
    -------
    data
        The encoded data.
    arr_lookup
        How to look up where to rebuild Numpy arrays.
    """

    arr_lookup: list[list] = []

    def encode(obj: Any, path: list) -> Any:
        if len(path) > MAX_DEPTH:
            raise RuntimeError("I'm in too deep!")

        if isinstance(obj, ZarrEncodable):
            return encode(obj.to_dict(), path)
        elif isinstance(obj, (list, tuple, set)):
            items = (encode(item, [*path, idx]) for idx, item in enumerate(obj))
            if isinstance(obj, tuple):
                return tuple(items)
            elif isinstance(obj, set):
                return set(items)
            else:
                return list(items)
        elif isinstance(obj, dict):
            return {key: encode(val, [*path, key]) for key, val in obj.items()}
        elif isinstance(obj, PRIMITIVES):
            return obj
        elif isinstance(obj, np.ndarray):
            # Might need to generate new group:
            param_arr_group = root_group.require_group(arr_path)
            new_idx = max((int(i) + 1 for i in param_arr_group.keys()), default=0)
            param_arr_group.create_dataset(name=f"arr_{new_idx}", data=obj)
            arr_lookup.append([path, new_idx])
            return None
        else:
            raise ValueError(f"unserializable type: {type(obj)}")

    return encode(obj, []), arr_lookup


def zarr_decode(
    param_data: None | dict,
    arr_group: zarr.Group,
    path: list | None = None,
    dataset_copy: bool = False,
):
    """
    Decode data from a zarr group.
    """
    if param_data is None:
        return None

    path = path or []

    data = get_in_container(param_data["data"], path)
    # data = copy.deepcopy(data)  # TODO: why did we copy?

    for arr_path, arr_idx in param_data["arr_lookup"]:
        try:
            rel_path = get_relative_path(arr_path, path)
        except ValueError:
            continue

        dataset = arr_group.get(f"arr_{arr_idx}")

        if dataset_copy:
            dataset = dataset[:]

        if rel_path:
            set_in_container(data, rel_path, dataset)
        else:
            data = dataset

    return data


class ZarrEncodable:
    """
    Base class of data that can be converted to and from zarr form.
    """

    _typ = None

    def to_dict(self) -> dict[str, Any]:
        """
        Convert this object to a dict.
        """
        if hasattr(self, "__dict__"):
            return dict(self.__dict__)
        elif hasattr(self, "__slots__"):
            return {k: getattr(self, k) for k in self.__slots__}
        else:
            # Should be unreachable
            return {}

    def to_zarr(self, zarr_group: zarr.Group):
        """
        Save this object into the given zarr group.
        """
        zarr_encode(self.to_dict(), zarr_group, is_pending_add=False, is_set=False)

    @classmethod
    def from_zarr(cls, zarr_group: zarr.Group, dataset_copy: bool = False) -> Self:
        """
        Read an instance of this class from the given zarr group.
        """
        # FIXME: Do the read of the data!
        param_data = None
        data = zarr_decode(param_data, zarr_group, dataset_copy=dataset_copy)
        return cls(**data)



"""
Utilities for making data to use in testing.
"""

from __future__ import annotations
from dataclasses import dataclass
from pathlib import Path
from typing import Any, ClassVar, TYPE_CHECKING
from hpcflow.app import app as hf
from hpcflow.sdk.core.parameters import ParameterValue
from hpcflow.sdk.core.utils import get_file_context
from hpcflow.sdk.submission.shells import ALL_SHELLS
from hpcflow.sdk.typing import hydrate


if TYPE_CHECKING:
    from collections.abc import Iterable, Mapping, Sequence
    from typing_extensions import TypeAlias, Self
    from h5py import Group as HDF5Group  # type: ignore
    from .actions import Action
    from .element import ElementGroup
    from .loop import Loop
    from .parameters import InputSource, Parameter, SchemaInput, InputValue
    from .task import Task
    from .task_schema import TaskSchema
    from .types import Resources
    from .workflow import Workflow, WorkflowTemplate
    from ..app import BaseApp
    from ..typing import PathLike
# mypy: disable-error-code="no-untyped-def"

#: A string or a tuple of strings.
Strs: TypeAlias = "str | tuple[str, ...]"


def make_schemas(
    *ins_outs: tuple[dict[str, Any], tuple[str, ...]]
    | tuple[dict[str, Any], tuple[str, ...], str]
    | tuple[dict[str, Any], tuple[str, ...], str, dict[str, Any]]
) -> list[TaskSchema]:
    """
    Construct a collection of schemas.
    """
    out: list[TaskSchema] = []
    for idx, info in enumerate(ins_outs):
        act_kwargs: dict[str, Any] = {}
        if len(info) == 2:
            (ins_i, outs_i) = info
            obj = f"t{idx}"
        elif len(info) == 3:
            (ins_i, outs_i, obj) = info
        else:
            (ins_i, outs_i, obj, act_kwargs) = info

        # distribute outputs over multiple commands' stdout:
        cmds_lst = []
        for out_idx, out_j in enumerate(outs_i):
            cmd = hf.Command(
                command=(
                    "echo $(("
                    + " + ".join(f"<<parameter:{i}>> + {100 + out_idx}" for i in ins_i)
                    + "))"
                ),
                stdout=f"<<int(parameter:{out_j})>>",
            )
            cmds_lst.append(cmd)

        if not outs_i:
            # no outputs
            cmds_lst = [
                hf.Command(
                    command=(
                        "echo $(("
                        + " + ".join(f"<<parameter:{i}>> + 100" for i in ins_i)
                        + "))"
                    ),
                )
            ]

        act_i = hf.Action(commands=cmds_lst, **act_kwargs)
        out.append(
            hf.TaskSchema(
                objective=obj,
                actions=[act_i],
                inputs=[hf.SchemaInput(k, default_value=v) for k, v in ins_i.items()],
                outputs=[hf.SchemaOutput(hf.Parameter(k)) for k in outs_i],
            )
        )
    return out


def make_parameters(num: int) -> list[Parameter]:
    """
    Construct a sequence of parameters.
    """
    return [hf.Parameter(f"p{i + 1}") for i in range(num)]


def make_actions(
    ins_outs: list[tuple[Strs, str] | tuple[Strs, str, str]],
    env: str = "env1",
) -> list[Action]:
    """
    Construct a collection of actions.
    """
    act_env = hf.ActionEnvironment(environment=env)
    actions = []
    for ins_outs_i in ins_outs:
        if len(ins_outs_i) == 2:
            ins, out = ins_outs_i
            err: str | None = None
        else:
            ins, out, err = ins_outs_i
        if not isinstance(ins, tuple):
            ins = (ins,)
        cmd_str = "doSomething "
        for i in ins:
            cmd_str += f" <<parameter:{i}>>"
        stdout = f"<<parameter:{out}>>"
        stderr = None
        if err:
            stderr = f"<<parameter:{err}>>"
        act = hf.Action(
            commands=[hf.Command(cmd_str, stdout=stdout, stderr=stderr)],
            environments=[act_env],
        )
        actions.append(act)
    return actions


def make_tasks(
    schemas_spec: Iterable[
        tuple[dict[str, Any], tuple[str, ...]]
        | tuple[dict[str, Any], tuple[str, ...], str]
    ],
    local_inputs: dict[int, Iterable[str]] | None = None,
    local_sequences: (
        dict[int, Iterable[tuple[str, int, int | float | None]]] | None
    ) = None,
    local_resources: dict[int, dict[str, dict]] | None = None,
    nesting_orders: dict[int, dict[str, float]] | None = None,
    input_sources: dict[int, dict[str, list[InputSource]]] | None = None,
    groups: dict[int, Iterable[ElementGroup]] | None = None,
) -> list[Task]:
    """
    Construct a sequence of tasks.
    """
    local_inputs = local_inputs or {}
    local_sequences = local_sequences or {}
    local_resources = local_resources or {}
    nesting_orders = nesting_orders or {}
    input_sources = input_sources or {}
    groups = groups or {}
    schemas = make_schemas(*schemas_spec)
    tasks: list[Task] = []
    for s_idx, s in enumerate(schemas):
        inputs = [
            hf.InputValue(hf.Parameter(i), value=int(i[1:]) * 100)
            for i in local_inputs.get(s_idx, ())
        ]
        seqs = [
            hf.ValueSequence(
                path=i[0],
                values=[(int(i[0].split(".")[1][1:]) * 100) + j for j in range(i[1])],
                nesting_order=i[2],
            )
            for i in local_sequences.get(s_idx, ())
        ]
        res = {k: v for k, v in local_resources.get(s_idx, {}).items()}
        task = hf.Task(
            schema=s,
            inputs=inputs,
            sequences=seqs,
            resources=res,
            nesting_order=nesting_orders.get(s_idx, {}),
            input_sources=input_sources.get(s_idx, None),
            groups=list(groups.get(s_idx, ())),
        )
        tasks.append(task)
    return tasks


def make_workflow(
    schemas_spec: Iterable[
        tuple[dict[str, Any], tuple[str, ...]]
        | tuple[dict[str, Any], tuple[str, ...], str]
    ],
    path: PathLike,
    local_inputs: dict[int, Iterable[str]] | None = None,
    local_sequences: (
        dict[int, Iterable[tuple[str, int, int | float | None]]] | None
    ) = None,
    local_resources: dict[int, dict[str, dict]] | None = None,
    nesting_orders: dict[int, dict[str, float]] | None = None,
    input_sources: dict[int, dict[str, list[InputSource]]] | None = None,
    resources: Resources = None,
    loops: list[Loop] | None = None,
    groups: dict[int, Iterable[ElementGroup]] | None = None,
    name: str = "w1",
    overwrite: bool = False,
    store: str = "zarr",
) -> Workflow:
    """
    Construct a workflow.
    """
    tasks = make_tasks(
        schemas_spec,
        local_inputs=local_inputs,
        local_sequences=local_sequences,
        local_resources=local_resources,
        nesting_orders=nesting_orders,
        input_sources=input_sources,
        groups=groups,
    )
    template: Mapping[str, Any] = {
        "name": name,
        "tasks": tasks,
        "resources": resources,
        **({"loops": loops} if loops else {}),
    }
    wk = hf.Workflow.from_template(
        hf.WorkflowTemplate(**template),
        path=path,
        name=name,
        overwrite=overwrite,
        store=store,
    )
    return wk


def make_test_data_YAML_workflow(
    workflow_name: str,
    path: PathLike,
    app: BaseApp | None = None,
    pkg: str = "hpcflow.tests.data",
    **kwargs,
) -> Workflow:
    """Generate a workflow whose template file is defined in the test data directory."""
    app = app or hf
    with get_file_context(pkg, workflow_name) as file_path:
        return app.Workflow.from_YAML_file(YAML_path=file_path, path=path, **kwargs)


def make_test_data_YAML_workflow_template(
    workflow_name: str,
    app: BaseApp | None = None,
    pkg: str = "hpcflow.tests.data",
    **kwargs,
) -> WorkflowTemplate:
    """Generate a workflow template whose file is defined in the test data directory."""
    app = app or hf
    with get_file_context(pkg, workflow_name) as file_path:
        return app.WorkflowTemplate.from_file(path=file_path, **kwargs)


@dataclass
@hydrate
class P1_sub_parameter_cls(ParameterValue):
    """
    Parameter value handler: ``p1_sub``
    """

    _typ: ClassVar[str] = "p1_sub"

    e: int = 0

    def CLI_format(self) -> str:
        return str(self.e)

    @property
    def twice_e(self):
        return self.e * 2

    def prepare_JSON_dump(self) -> dict[str, Any]:
        return {"e": self.e}

    def dump_to_HDF5_group(self, group: HDF5Group):
        group.attrs["e"] = self.e


@dataclass
@hydrate
class P1_sub_parameter_cls_2(ParameterValue):
    """
    Parameter value handler: ``p1_sub_2``
    """

    _typ: ClassVar[str] = "p1_sub_2"

    f: int = 0


@dataclass
@hydrate
class P1_parameter_cls(ParameterValue):
    """
    Parameter value handler: ``p1c``

    Note
    ----
    This is a composite value handler.
    """

    _typ: ClassVar[str] = "p1c"
    _sub_parameters: ClassVar[dict[str, str]] = {
        "sub_param": "p1_sub",
        "sub_param_2": "p1_sub_2",
    }

    a: int = 0
    d: int | None = None
    sub_param: P1_sub_parameter_cls | None = None

    def __post_init__(self):
        if self.sub_param is not None and not isinstance(
            self.sub_param, P1_sub_parameter_cls
        ):
            self.sub_param = P1_sub_parameter_cls(**self.sub_param)

    @classmethod
    def from_data(cls, b: int, c: int):
        return cls(a=b + c)

    @classmethod
    def from_file(cls, path: str):
        with Path(path).open("rt") as fh:
            lns = fh.readlines()
            a = int(lns[0])
        return cls(a=a)

    @property
    def twice_a(self) -> int:
        return self.a * 2

    @property
    def sub_param_prop(self) -> P1_sub_parameter_cls:
        return P1_sub_parameter_cls(e=4 * self.a)

    def CLI_format(self) -> str:
        return str(self.a)

    @staticmethod
    def CLI_format_group(*objs) -> str:
        return ""

    @staticmethod
    def sum(*objs, **kwargs) -> str:
        return str(sum(i.a for i in objs))

    def custom_CLI_format(self, add: str | None = None, sub: str | None = None) -> str:
        add_i = 4 if add is None else int(add)
        sub_i = 0 if sub is None else int(sub)
        return str(self.a + add_i - sub_i)

    def custom_CLI_format_prep(self, reps: str | None = None) -> list[int]:
        """Used for testing custom object CLI formatting.

        For example, with a command like this:

        `<<join[delim=","](parameter:p1c.custom_CLI_format_prep(reps=4))>>`.

        """
        reps_int = 1 if reps is None else int(reps)
        return [self.a] * reps_int

    @classmethod
    def CLI_parse(cls, a_str: str, double: str = "", e: str | None = None):
        a = int(a_str)
        if double.lower() == "true":
            a *= 2
        if e:
            sub_param = P1_sub_parameter_cls(e=int(e))
        else:
            sub_param = None
        return cls(a=a, sub_param=sub_param)

    def prepare_JSON_dump(self) -> dict[str, Any]:
        sub_param_js = self.sub_param.prepare_JSON_dump() if self.sub_param else None
        return {"a": self.a, "d": self.d, "sub_param": sub_param_js}

    def dump_to_HDF5_group(self, group: HDF5Group):
        group.attrs["a"] = self.a
        if self.d is not None:
            group.attrs["d"] = self.d
        if self.sub_param:
            sub_group = group.create_group("sub_param")
            self.sub_param.dump_to_HDF5_group(sub_group)

    @classmethod
    def dump_element_group_to_HDF5_group(self, objs: list[Self], group: HDF5Group):
        """
        Write a list (from an element group) of parameter values to an HDF5 group.
        """

        for obj_idx, p1_obj in enumerate(objs):
            grp_i = group.create_group(f"{obj_idx}")
            grp_i.attrs["a"] = p1_obj.a
            if p1_obj.d is not None:
                group.attrs["d"] = p1_obj.d
            if p1_obj.sub_param:
                sub_group = grp_i.create_group("sub_param")
                p1_obj.sub_param.dump_to_HDF5_group(sub_group)

    @classmethod
    def save_from_JSON(cls, data: dict, param_id: int | list[int], workflow: Workflow):
        obj = cls(**data)  # TODO: pass sub-param
        workflow.set_parameter_value(param_id=param_id, value=obj, commit=True)

    @classmethod
    def save_from_HDF5_group(cls, group: HDF5Group, param_id: int, workflow: Workflow):
        a = group.attrs["a"].item()
        if "d" in group.attrs:
            d = group.attrs["d"].item()
        else:
            d = None
        if "sub_param" in group:
            sub_group = group.get("sub_param")
            e = sub_group.attrs["e"].item()
            sub_param = P1_sub_parameter_cls(e=e)
        else:
            sub_param = None
        obj = cls(a=a, d=d, sub_param=sub_param)
        workflow.set_parameter_value(param_id=param_id, value=obj, commit=True)


def make_workflow_to_run_command(
    command,
    path,
    outputs=None,
    name="w1",
    overwrite=False,
    store="zarr",
    requires_dir=False,
):
    """Generate a single-task single-action workflow that runs the specified command,
    optionally generating some outputs."""

    outputs = outputs or []
    commands = [hf.Command(command=command)]
    commands += [
        hf.Command(command=f'echo "output_{out}"', stdout=f"<<parameter:{out}>>")
        for out in outputs
    ]
    schema = hf.TaskSchema(
        objective="run_command",
        outputs=[hf.SchemaOutput(i) for i in outputs],
        actions=[hf.Action(commands=commands, requires_dir=requires_dir)],
    )
    template = {
        "name": name,
        "tasks": [hf.Task(schema=schema)],
    }
    wk = hf.Workflow.from_template(
        hf.WorkflowTemplate(**template),
        path=path,
        name=name,
        overwrite=overwrite,
        store=store,
    )
    return wk


def command_line_test(
    cmd_str: str,
    expected: str,
    inputs: dict[str, Any] | list[InputValue],
    path: Path,
    outputs: Sequence[str] | None = None,
    cmd_stdout: str | None = None,
    shell_args: tuple[str, str] | None = None,
    schema_inputs: list[Parameter | SchemaInput] | None = None,
):
    """Utility function for testing `Command.get_command_line` in various scenarios, via
    a single-action, single-command workflow.

    The functions asserts that the generated command line based on `cmd_str` is equal to
    the provided `expected` command line.

    Parameters
    ----------
    cmd_str
        The command string to test.
    expected
        The resolved commandline string that should be generated.
    inputs
        Either a dictionary mapping string input names to values, or a list of
        `InputValue` objects.
    path
        The path to use to create the workflow during the test.
    outputs
        List of string output names.
    cmd_stdout
        The `Command` object's stdout attribute.
    shell_args
        Tuple of shell name and os name, used to select which `Shell` to instantiate.
    schema_inputs
        List of `SchemaInput` objects to use. If not passed, simple schema inputs will be
        generated.

    """

    inputs_ = (
        [hf.InputValue(inp_name, value=inp_val) for inp_name, inp_val in inputs.items()]
        if isinstance(inputs, dict)
        else inputs
    )

    schema_inputs_ = (
        schema_inputs
        if schema_inputs
        else [hf.SchemaInput(parameter=inp_val.parameter) for inp_val in inputs_]
    )

    s1 = hf.TaskSchema(
        objective="t1",
        inputs=schema_inputs_,
        outputs=[
            hf.SchemaOutput(parameter=hf.Parameter(out_name))
            for out_name in outputs or ()
        ],
        actions=[hf.Action(commands=[hf.Command(command=cmd_str, stdout=cmd_stdout)])],
    )
    wk = hf.Workflow.from_template_data(
        tasks=[hf.Task(schema=s1, inputs=inputs_)],
        path=path,
        template_name="test_get_command_line",
        overwrite=True,
    )
    t1 = wk.tasks.t1
    assert isinstance(t1, hf.WorkflowTask)
    run = t1.elements[0].iterations[0].action_runs[0]
    command = run.action.commands[0]
    shell_args_ = shell_args or ("powershell", "nt")
    shell = ALL_SHELLS[shell_args_[0]][shell_args_[1]]()
    cmd_line, _ = command.get_command_line(
        EAR=run, shell=shell, env=run.get_environment()
    )
    assert cmd_line == expected



"""
Model of a command run in an action.
"""

from __future__ import annotations
from dataclasses import dataclass, field
from functools import partial
from pathlib import Path
import re
from typing import Any, ClassVar, TYPE_CHECKING

import numpy as np

from hpcflow.sdk.log import TimeIt
from hpcflow.sdk.typing import hydrate
from hpcflow.sdk.core.element import ElementResources
from hpcflow.sdk.core.errors import NoCLIFormatMethodError
from hpcflow.sdk.core.json_like import ChildObjectSpec, JSONLike
from hpcflow.sdk.core.parameters import ParameterValue

if TYPE_CHECKING:
    from collections.abc import Callable, Iterable, Mapping, Sequence
    from re import Pattern
    from .actions import ActionRule, Action
    from .element import ElementActionRun
    from .environment import Environment
    from ..submission.shells import Shell


@dataclass
@hydrate
class Command(JSONLike):
    """
    A command that may be run within a workflow action.

    Parameters
    ----------
    command: str
        The actual command.
    executable: str
        The executable to run,
        from the set of executable managed by the environment.
    arguments: list[str]
        The arguments to pass in.
    variables: dict[str, str]
        Values that may be substituted when preparing the arguments.
    stdout: str
        The name of a file to write standard output to.
    stderr: str
        The name of a file to write standard error to.
    stdin: str
        The name of a file to read standard input from.
    rules: list[~hpcflow.app.ActionRule]
        Rules that state whether this command is eligible to run.
    """

    _child_objects: ClassVar[tuple[ChildObjectSpec, ...]] = (
        ChildObjectSpec(
            name="rules",
            class_name="ActionRule",
            is_multiple=True,
            parent_ref="command",
        ),
    )

    #: The actual command.
    #: Overrides :py:attr:`executable`.
    command: str | None = None
    #: The executable to run,
    #: from the set of executable managed by the environment.
    executable: str | None = None
    #: The arguments to pass in.
    arguments: list[str] | None = None
    #: Values that may be substituted when preparing the arguments.
    variables: dict[str, str] | None = None
    #: The name of a file to write standard output to.
    stdout: str | None = None
    #: The name of a file to write standard error to.
    stderr: str | None = None
    #: The name of a file to read standard input from.
    stdin: str | None = None
    #: Rules that state whether this command is eligible to run.
    rules: list[ActionRule] = field(default_factory=list)

    action: Action | None = None  # assigned by parent Action

    def __post_init__(self):
        self._set_parent_refs()

    def __repr__(self) -> str:
        out = []
        if self.command:
            out.append(f"command={self.command!r}")
        if self.executable:
            out.append(f"executable={self.executable!r}")
        if self.arguments:
            out.append(f"arguments={self.arguments!r}")
        if self.variables:
            out.append(f"variables={self.variables!r}")
        if self.stdout:
            out.append(f"stdout={self.stdout!r}")
        if self.stderr:
            out.append(f"stderr={self.stderr!r}")
        if self.stdin:
            out.append(f"stdin={self.stdin!r}")
        if self.rules:
            out.append(f"rules={self.rules!r}")

        return f"{self.__class__.__name__}({', '.join(out)})"

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return (
            self.command == other.command
            and self.executable == other.executable
            and self.arguments == other.arguments
            and self.variables == other.variables
            and self.stdout == other.stdout
            and self.stderr == other.stderr
            and self.stdin == other.stdin
            and self.rules == other.rules
        )

    def __get_initial_command_line(self) -> str:
        if self.command:
            return self.command
        else:
            return self.executable or ""

    __EXE_SCRIPT_RE: ClassVar[Pattern] = re.compile(r"\<\<(executable|script):(.*?)\>\>")
    __ENV_SPEC_RE: ClassVar[Pattern] = re.compile(r"\<\<env:(.*?)\>\>")

    def get_command_line(
        self, EAR: ElementActionRun, shell: Shell, env: Environment
    ) -> tuple[str, list[tuple[str, ...]]]:
        """Return the resolved command line.

        This is ordinarily called at run-time by `Workflow.write_commands`.
        """

        self._app.persistence_logger.debug("Command.get_command_line")
        cmd_str = self.__get_initial_command_line()

        def _format_sum(iterable: Iterable) -> str:
            return str(sum(iterable))

        def _join(iterable: Iterable, delim: str) -> str:
            return delim.join(map(str, iterable))

        parse_types: dict[str, Callable[..., str]] = {
            "sum": _format_sum,
            "join": _join,
        }

        def exec_script_repl(match_obj: re.Match[str]) -> str:
            typ, val = match_obj.groups()
            if typ == "executable":
                executable = env.executables.get(val)
                filterable = ElementResources.get_env_instance_filterable_attributes()
                filter_exec = {attr: EAR.get_resources().get(attr) for attr in filterable}
                exec_cmd = executable.filter_instances(**filter_exec)[0].command
                return exec_cmd.replace("<<num_cores>>", str(EAR.resources.num_cores))
            elif typ == "script":
                # TODO: is this needed? we have <<script_name>> <<script_path>> etc as command variables
                return EAR.action.get_script_name(val)
            else:
                raise ValueError("impossible match occurred")

        def input_param_repl(match_obj: re.Match[str], inp_val) -> str:
            _, func, func_kwargs, method, method_kwargs = match_obj.groups()

            if isinstance(inp_val, ParameterValue):
                if not method:
                    method = "CLI_format"
                if not hasattr(inp_val, method):
                    raise NoCLIFormatMethodError(method, inp_val)
                kwargs = self.__prepare_kwargs_from_string(args_str=method_kwargs)
                inp_val = getattr(inp_val, method)(**kwargs)

            if func:
                kwargs = self.__prepare_kwargs_from_string(
                    args_str=func_kwargs,
                    doubled_quoted_args=["delim"],
                )
                inp_val = parse_types[func](inp_val, **kwargs)

            return str(inp_val)

        file_regex = r"(\<\<file:{}\>\>?)"

        # substitute executables:
        cmd_str = self.__EXE_SCRIPT_RE.sub(
            repl=exec_script_repl,
            string=cmd_str,
        )

        # executable command might itself contain variables defined in `variables`, and/or
        # an `<<args>>` variable::
        for var_key, var_val in (self.variables or {}).items():
            # substitute any `<<env:>>` specifiers
            var_val = self.__ENV_SPEC_RE.sub(
                repl=lambda match_obj: EAR.env_spec[match_obj[1]],
                string=var_val,
            )
            cmd_str = cmd_str.replace(f"<<{var_key}>>", var_val)
            if "<<args>>" in cmd_str:
                args_str = " ".join(self.arguments or ())
                ends_in_args = cmd_str.endswith("<<args>>")
                cmd_str = cmd_str.replace("<<args>>", args_str)
                if ends_in_args and not args_str:
                    cmd_str = cmd_str.rstrip()

        # remove any left over "<<args>>" and "<<script_name>>"s:
        cmd_str = (
            cmd_str.replace("<<args>>", "")
            .replace("<<script_name>>", "")
            .replace("<<script_path>>", "")
        )

        # substitute input parameters in command:
        types_pattern = "|".join(parse_types)
        pattern = (
            r"(\<\<(?:({types_pattern})(?:\[(.*)\])?\()?parameter:{name}(?:\.(\w+)"
            r"\((.*?)\))?\)?\>\>?)"
        )

        for cmd_inp_full in EAR.action.get_command_input_types(sub_parameters=True):
            # remove any CLI formatting method, which will be the final component and will
            # include parentheses:
            cmd_inp_parts = cmd_inp_full.split(".")
            if "(" in cmd_inp_parts[-1]:
                cmd_inp = ".".join(cmd_inp_parts[:-1])
            else:
                cmd_inp = cmd_inp_full
            inp_val = EAR.get(
                f"inputs.{cmd_inp}",
                raise_on_unset=True,
            )  # TODO: what if schema output?
            pattern_i = pattern.format(
                types_pattern=types_pattern,
                name=re.escape(cmd_inp),
            )
            cmd_str = re.sub(
                pattern=pattern_i,
                repl=partial(input_param_repl, inp_val=inp_val),
                string=cmd_str,
            )

        # substitute input/output files in command:
        for cmd_file in EAR.action.get_command_file_labels():
            file_path = EAR.get(
                f"input_files.{cmd_file}", raise_on_unset=True
            ) or EAR.get(f"output_files.{cmd_file}", raise_on_unset=True)
            # assuming we have copied this file to the EAR directory, then we just
            # need the file name:
            file_name = Path(file_path).name
            cmd_str = re.sub(
                pattern=file_regex.format(cmd_file),
                repl=file_name,
                string=cmd_str,
            )

        shell_vars: list[tuple[str, ...]] = []
        out_types = self.get_output_types()
        if out_types["stdout"]:
            # TODO: also map stderr/both if possible
            # assign stdout to a shell variable if required:
            param_name = f"outputs.{out_types['stdout']}"
            shell_var_name = f"parameter_{out_types['stdout']}"
            shell_vars.append((param_name, shell_var_name, "stdout"))
            cmd_str = shell.format_stream_assignment(
                shell_var_name=shell_var_name,
                command=cmd_str,
            )
        elif self.stdout:
            cmd_str += f" 1>> {self.stdout}"

        if self.stderr:
            cmd_str += f" 2>> {self.stderr}"

        return cmd_str, shell_vars

    # note: we use "parameter" rather than "output", because it could be a schema
    # output or schema input.
    __PARAM_RE: ClassVar[Pattern] = re.compile(
        r"(?:\<\<(?:\w+(?:\[(?:.*)\])?\()?parameter:(\w+)"
        r"(?:\.(?:\w+)\((?:.*?)\))?\)?\>\>?)"
    )

    def get_output_types(self) -> Mapping[str, str | None]:
        """
        Get whether stdout and stderr are workflow parameters.
        """
        out: dict[str, str | None] = {"stdout": None, "stderr": None}
        for i, label in zip((self.stdout, self.stderr), ("stdout", "stderr")):
            if i and (match := self.__PARAM_RE.search(i)):
                param_typ: str = match[1]
                if match.span(0) != (0, len(i)):
                    raise ValueError(
                        f"If specified as a parameter, `{label}` must not include"
                        f" any characters other than the parameter "
                        f"specification, but this was given: {i!r}."
                    )
                out[label] = param_typ
        return out

    @staticmethod
    def __prepare_kwargs_from_string(
        args_str: str | None, doubled_quoted_args: list[str] | None = None
    ) -> dict[str, str]:
        if args_str is None:
            return {}

        kwargs: dict[str, str] = {}
        # deal with specified double-quoted arguments first if it exists:
        for quote_arg in doubled_quoted_args or ():
            quote_pat = r'.*({quote_arg}="(.*)").*'.format(quote_arg=quote_arg)
            if match := re.match(quote_pat, args_str):
                quote_str, quote_contents = match.groups()
                args_str = args_str.replace(quote_str, "")
                kwargs[quote_arg] = quote_contents

        if args_str := args_str.strip().strip(","):
            for arg_part in args_str.split(","):
                name_i, value_i = map(str.strip, arg_part.split("="))
                kwargs[name_i] = value_i
        return kwargs

    def process_std_stream(self, name: str, value: str, stderr: bool) -> Any:
        """
        Process a description of a standard stread from a command to get how it becomes
        a workflow parameter for later actions.

        Parameters
        ---------
        name:
            The name of the output, describing how to process things.
        value:
            The actual value read from the stream.
        stderr:
            If true, this is handling the stderr stream. If false, the stdout stream.
        """

        def _parse_list(
            lst_str: str, item_type: str = "str", delim: str = " "
        ) -> list[Any]:
            return [parse_types[item_type](i) for i in lst_str.split(delim)]

        def _parse_array(
            arr_str: str, item_type: str = "float", delim: str = " "
        ) -> np.ndarray[Any, np.dtype[Any]]:
            return np.array(
                _parse_list(lst_str=arr_str, item_type=item_type, delim=delim)
            )

        def _parse_bool(bool_str: str) -> bool:
            bool_str = bool_str.lower()
            if bool_str in ("true", "1"):
                return True
            elif bool_str in ("false", "0"):
                return False
            else:
                raise ValueError(
                    f"Cannot parse value {bool_str!r} as a boolean in command "
                    f"{'stderr' if stderr else 'stdout'}: "
                    f"{self.stderr if stderr else self.stdout!r}."
                )

        parse_types: dict[str, Callable[[str], Any]] = {
            "str": str,
            "int": int,
            "float": float,
            "bool": _parse_bool,
            "list": _parse_list,
            "array": _parse_array,
        }
        types_pattern = "|".join(parse_types)

        # TODO: use str.removeprefix in 3.9 onwards
        out_name = name.replace("outputs.", "")
        pattern = (
            r"(\<\<(?:({types_pattern})(?:\[(.*)\])?\()?parameter:{name}(?:\.(\w+)"
            r"\((.*?)\))?\)?\>\>?)"
        )
        pattern = pattern.format(types_pattern=types_pattern, name=out_name)
        spec = self.stderr if stderr else self.stdout
        assert spec is not None
        self._app.submission_logger.info(
            f"processing shell standard stream according to spec: {spec!r}"
        )
        param = self._app.Parameter(out_name)
        if (match := re.match(pattern, spec)) is None:
            return value
        groups = match.groups()
        parse_type, parse_args_str = groups[1:3]
        parse_args = self.__prepare_kwargs_from_string(
            args_str=parse_args_str,
            doubled_quoted_args=["delim"],
        )
        if param._value_class:
            method, method_args_str = groups[3:5]
            method_args = self.__prepare_kwargs_from_string(
                args_str=method_args_str,
                doubled_quoted_args=["delim"],
            )
            method = method or "CLI_parse"
            value = getattr(param._value_class, method)(value, **method_args)
        if parse_type:
            value = parse_types[parse_type](value, **parse_args)

        return value

    __EXE_RE: ClassVar[Pattern] = re.compile(r"\<\<(?:executable):(.*?)\>\>")

    @classmethod
    def _extract_executable_labels(cls, cmd_str: str) -> Sequence[str]:
        return cls.__EXE_RE.findall(cmd_str)

    @TimeIt.decorator
    def get_required_executables(self) -> Sequence[str]:
        """Return executable labels required by this command."""
        # an executable label might appear in the `command` or `executable` attribute:
        return self._extract_executable_labels(self.__get_initial_command_line())



"""
Class to hold the state that is waiting to be committed to disk.
"""

from __future__ import annotations

from collections import defaultdict
from collections.abc import Callable
import contextlib
from dataclasses import dataclass, field, fields
from functools import wraps
import copy

from typing import Any, Generic, TYPE_CHECKING, TypeVar
from typing_extensions import ParamSpec

import numpy as np

from hpcflow.sdk.log import TimeIt
from hpcflow.sdk.submission.submission import SUBMISSION_SUBMIT_TIME_KEYS
from hpcflow.sdk.typing import DataIndex
from hpcflow.sdk.utils.strings import shorten_list_str


from hpcflow.sdk.persistence.types import (
    AnySTask,
    AnySElement,
    AnySElementIter,
    AnySEAR,
    AnySParameter,
)

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence
    from datetime import datetime
    from logging import Logger
    from .base import PersistentStore, FileDescriptor, LoopDescriptor
    from ..app import BaseApp
    from ..typing import ParamSource
    from ..core.json_like import JSONed

P = ParamSpec("P")
T = TypeVar("T")

_commit_method_data_map: dict[str, list[str]] = defaultdict(
    list
)  # note: this is updated at module-import time.


class PendingChanges(
    Generic[AnySTask, AnySElement, AnySElementIter, AnySEAR, AnySParameter]
):
    """
    Class to store pending changes and merge them into a persistent store.

    Parameters
    ----------
    app: App
        The main application context.
    store: PersistentStore
        The persistent store that owns this object
    resource_map: CommitResourceMap
        Map of resources, used when processing commits.
    """

    # These would be in the docstring except they render really wrongly!
    # Type Parameters
    # ---------------
    # AnySTask
    #     The type of stored tasks.
    # AnySElement
    #     The type of stored elements.
    # AnySElementIter
    #     The type of stored element iterations.
    # AnySEAR
    #     The type of stored EARs.
    # AnySParameter
    #     The type of stored parameters.

    def __init__(
        self,
        app: BaseApp,
        store: PersistentStore[
            AnySTask, AnySElement, AnySElementIter, AnySEAR, AnySParameter
        ],
        resource_map: CommitResourceMap,
    ):
        self._app = app
        self.store = store
        self.resource_map = resource_map

        #: Keys are new task IDs.
        self.add_tasks: dict[int, AnySTask] = {}
        #: Keys are loop IDs, values are loop descriptors.
        self.add_loops: dict[int, LoopDescriptor] = {}
        #: Keys are submission IDs, values are submission descriptors.
        self.add_submissions: dict[int, Mapping[str, JSONed]] = {}
        #: Keys are element IDs.
        self.add_elements: dict[int, AnySElement] = {}
        #: Keys are element iteration IDs.
        self.add_elem_iters: dict[int, AnySElementIter] = {}
        #: Keys are element action run IDs.
        self.add_EARs: dict[int, AnySEAR] = {}
        #: Keys are parameter indices and values are tuples whose first element is data
        #: to add and whose second element is the source dict for the new data.
        self.add_parameters: dict[int, AnySParameter] = {}
        #: Workflow-related files (inputs, outputs) added to the persistent store.
        self.add_files: list[FileDescriptor] = []
        #: Template components to add.
        self.add_template_components: dict[str, dict[str, dict]] = {}
        #: Keys are element set IDs, values are descriptors.
        self.add_element_sets: dict[int, list[Mapping]] = {}

        #: Keys are task IDs, and values are element IDs to add to that task.
        self.add_elem_IDs: dict[int, list[int]] = {}
        #: Keys are element IDs, and values are iteration IDs to add to that element.
        self.add_elem_iter_IDs: dict[int, list[int]] = {}
        #: Keys are element iteration IDs, then EAR action index, and values are EAR IDs.
        #: This is a list of EAR IDs to add to a given element iteration action.
        self.add_elem_iter_EAR_IDs: dict[int, dict[int, list[int]]] = {}
        #: Submission metadata added at submit-time, including submission parts.
        self.update_at_submit_metadata: dict[int, dict[str, Any]] = {}

        #: IDs of EARs to mark as initialised.
        self.set_EARs_initialised: list[int] = []
        #: Submission IDs and commands file IDs to attach to EARs.
        self.set_EAR_submission_data: dict[int, tuple[int, int | None]] = {}
        #: IDs of EARs to mark as skipped.
        self.set_EAR_skips: dict[int, int] = {}
        #: Keys are EAR IDs and values are tuples of start time, start dir snapshot, run
        #: hostname, and port number.
        self.set_EAR_starts: dict[
            int, tuple[datetime, dict[str, Any] | None, str, int | None]
        ] = {}
        #: Keys are EAR IDs and values are tuples of end time, end dir snapshot, exit
        #: code, and success boolean.
        self.set_EAR_ends: dict[
            int, tuple[datetime, dict[str, Any] | None, int, bool]
        ] = {}
        #: Each list item is a tuple of two arrays, the first of which is a run directory
        #: indices array, and the second of which is an integer array indicating with
        #: which run ID each run directory is associated.
        self.set_run_dirs: list[tuple[np.ndarray, np.ndarray]] = []

        #: Keys are IDs of jobscripts.
        self.set_js_metadata: dict[int, dict[int, dict[str, Any]]] = {}

        #: Keys are IDs of parameters to add or modify, and values are tuples of the
        #: parameter value, and whether the parameter is a file.
        self.set_parameters: dict[int, tuple[Any, bool]] = {}

        #: Keys are parameter indices and values are dict parameter sources to merge
        #: with existing source of that parameter.
        self.update_param_sources: dict[int, ParamSource] = {}
        #: Keys are indices of loops, values are descriptions of what to update.
        self.update_loop_indices: dict[int, dict[str, int]] = {}
        #: Keys are indices of loops, values are number of iterations.
        self.update_loop_num_iters: dict[int, list[list[list[int] | int]]] = {}
        #: Keys are indices of loops, values are list of parent names.
        self.update_loop_parents: dict[int, list[str]] = {}

        self.update_iter_data_idx: dict[int, DataIndex] = {}
        self.update_run_data_idx: dict[int, DataIndex] = {}

        self.reset(is_init=True)  # set up initial data structures

    def __bool__(self):
        """Returns True if there are any outstanding pending items."""
        return (
            bool(self.add_tasks)
            or bool(self.add_loops)
            or bool(self.add_submissions)
            or bool(self.add_elements)
            or bool(self.add_elem_iters)
            or bool(self.add_EARs)
            or bool(self.add_elem_IDs)
            or bool(self.add_elem_iter_IDs)
            or bool(self.add_elem_iter_EAR_IDs)
            or bool(self.update_at_submit_metadata)
            or bool(self.add_parameters)
            or bool(self.add_files)
            or bool(self.add_template_components)
            or bool(self.add_element_sets)
            or bool(self.set_EARs_initialised)
            or bool(self.set_EAR_submission_data)
            or bool(self.set_EAR_starts)
            or bool(self.set_EAR_ends)
            or bool(self.set_EAR_skips)
            or bool(self.set_run_dirs)
            or bool(self.set_js_metadata)
            or bool(self.set_parameters)
            or bool(self.update_param_sources)
            or bool(self.update_loop_indices)
            or bool(self.update_loop_num_iters)
            or bool(self.update_loop_parents)
            or bool(self.update_iter_data_idx)
            or bool(self.update_run_data_idx)
        )

    def where_pending(self) -> list[str]:
        """
        Get the list of items for which there is some outstanding pending items.
        """
        excluded = {"app", "store", "resource_map"}
        return [k for k, v in self.__dict__.items() if k not in excluded and bool(v)]

    @property
    def logger(self) -> Logger:
        """
        The logger.
        """
        return self._app.persistence_logger

    def commits_data(*data_list: str):
        """Decorator that wraps `PendingChanges.commit_*` methods with arguments listing
        which `PendingChanges` attributes must have non-trivial data in them for the method's
        invocation to be required.

        Notes
        -----
        This essentially provides a mapping between `PendingChanges` attributes and
        `commit_*` methods. This allows us to only open the resources that need updating
        in `PendingChanges.commit_all`.

        We use a decorator rather than an explicitly declaring the map in
        `_commit_method_data_map` to make the mapping obvious near the commit methods, and
        hopefully avoid us forgetting to update `_commit_method_data_map` when we modify
        or add commit methods in future!

        """

        def decorator(func: Callable[P, T]) -> Callable[P, T]:

            _commit_method_data_map[func.__name__].extend(data_list)

            @wraps(func)
            def inner(*args, **kwargs) -> T:
                return func(*args, **kwargs)

            return inner

        return decorator

    def get_pending_resource_map_groups(self) -> dict[tuple[str, ...], list[str]]:
        """Retrive resource map groups, where values are filtered to include only those
        commit methods that must be invoked, due to pending data associated with those
        methods.

        Notes
        -----
        This method allows us to open only those resources that need to be updated, given
        the state of pending data.
        """

        where_pending = self.where_pending()
        pending_groups = {}
        for res_names, methods in self.resource_map.groups.items():
            req_methods = [
                meth_i
                for meth_i in methods
                if any(
                    dat_j in where_pending for dat_j in _commit_method_data_map[meth_i]
                )
            ]
            if req_methods:
                pending_groups[res_names] = req_methods

        return pending_groups

    @TimeIt.decorator
    def commit_all(self) -> None:
        """Commit all pending changes to disk."""
        self.logger.info(f"committing all pending changes: {self.where_pending()}")

        if not self:
            self.logger.debug("commit: no pending changes to commit.")
            return

        for resources, methods in self.get_pending_resource_map_groups().items():
            # for each resource, enter `using_resource` context manager in "update" mode:
            with contextlib.ExitStack() as stack:
                for res in resources:
                    stack.enter_context(
                        self.store.using_resource(res, "update")  # type: ignore[call-overload]
                    )
                for meth in methods:
                    getattr(self, meth)()

        assert not (self)

    @TimeIt.decorator
    @commits_data("add_tasks")
    def commit_tasks(self) -> None:
        """Commit pending tasks to disk."""
        if self.add_tasks:
            tasks = self.store.get_tasks_by_IDs(self.add_tasks)
            task_ids = set(self.add_tasks)
            self.logger.debug(f"commit: adding pending tasks with IDs: {task_ids!r}")
            self.store._append_tasks(tasks)
            self.store.num_tasks_cache = None  # invalidate cache
            # pending element IDs that belong to pending tasks are now committed:
            self.add_elem_IDs = {
                k: v for k, v in self.add_elem_IDs.items() if k not in task_ids
            }
        self._clear_add_tasks()

    @TimeIt.decorator
    @commits_data("add_loops")
    def commit_loops(self) -> None:
        """Commit pending loops to disk."""
        if self.add_loops:
            # retrieve pending loops, including pending changes to num_added_iterations:
            loops = self.store.get_loops_by_IDs(self.add_loops)
            loop_ids = set(self.add_loops)
            self.logger.debug(f"commit: adding pending loops with indices {loop_ids!r}")
            self.store._append_loops(loops)

            # pending num_added_iters and parents that belong to pending loops are now
            # committed:
            self.update_loop_num_iters = {
                k: v for k, v in self.update_loop_num_iters.items() if k not in loop_ids
            }
            self.update_loop_parents = {
                k: v for k, v in self.update_loop_parents.items() if k not in loop_ids
            }

        self._clear_add_loops()

    @TimeIt.decorator
    @commits_data("add_submissions")
    def commit_submissions(self) -> None:
        """Commit pending submissions to disk."""
        if self.add_submissions:
            # retrieve pending submissions:
            subs = self.store.get_submissions_by_ID(
                self.add_submissions
            )  # TODO: I think this just returns add_submissions?
            sub_ids = set(self.add_submissions)
            self.logger.debug(
                f"commit: adding pending submissions with indices {sub_ids!r}"
            )
            self.store._append_submissions(subs)
        self._clear_add_submissions()

    @TimeIt.decorator
    @commits_data("update_at_submit_metadata")
    def commit_at_submit_metadata(self) -> None:
        """
        Commit to disk pending at-submit-time metadata, including submission parts.
        """
        if self.update_at_submit_metadata:
            self.logger.debug("commit: adding pending at-submit metadata")
            self.store._update_at_submit_metadata(self.update_at_submit_metadata)
        self._clear_at_submit_metadata()

    @TimeIt.decorator
    @commits_data("add_elem_IDs")
    def commit_elem_IDs(self) -> None:
        """
        Commit pending element ID updates to disk.
        """
        # TODO: could be batched up?
        for task_ID, elem_IDs in self.add_elem_IDs.items():
            self.logger.debug(
                f"commit: adding pending element IDs to task {task_ID!r}: {elem_IDs!r}."
            )
            self.store._append_task_element_IDs(task_ID, elem_IDs)
            self.store.task_cache.pop(task_ID, None)  # invalidate cache
        self._clear_add_elem_IDs()

    @TimeIt.decorator
    @commits_data("add_elements")
    def commit_elements(self) -> None:
        """
        Commit pending elements to disk.
        """
        if self.add_elements:
            elems = self.store.get_elements(self.add_elements)
            elem_ids = set(self.add_elements)
            self.logger.debug(f"commit: adding pending elements with IDs: {elem_ids!r}")
            self.store._append_elements(elems)
            # pending iter IDs that belong to pending elements are now committed:
            self.add_elem_iter_IDs = {
                k: v for k, v in self.add_elem_iter_IDs.items() if k not in elem_ids
            }
        self._clear_add_elements()

    @TimeIt.decorator
    @commits_data("add_element_sets")
    def commit_element_sets(self) -> None:
        """
        Commit pending element sets to disk.
        """
        # TODO: could be batched up?
        for task_id, es_js in self.add_element_sets.items():
            self.logger.debug("commit: adding pending element sets.")
            self.store._append_element_sets(task_id, es_js)
        self._clear_add_element_sets()

    @TimeIt.decorator
    @commits_data("add_elem_iter_IDs")
    def commit_elem_iter_IDs(self) -> None:
        """
        Commit pending element iteration ID updates to disk.
        """
        # TODO: could be batched up?
        for elem_ID, iter_IDs in self.add_elem_iter_IDs.items():
            self.logger.debug(
                f"commit: adding pending element iteration IDs to element {elem_ID!r}: "
                f"{iter_IDs!r}."
            )
            self.store._append_elem_iter_IDs(elem_ID, iter_IDs)
            self.store.element_cache.pop(elem_ID, None)  # invalidate cache
        self._clear_add_elem_iter_IDs()

    @TimeIt.decorator
    @commits_data("add_elem_iters")
    def commit_elem_iters(self) -> None:
        """
        Commit pending element iterations to disk.
        """
        if self.add_elem_iters:
            iters = self.store.get_element_iterations(self.add_elem_iters)
            iter_ids = set(self.add_elem_iters)
            self.logger.debug(
                f"commit: adding pending element iterations with IDs: {iter_ids!r}"
            )
            self.store._append_elem_iters(iters)
            # pending EAR IDs that belong to pending iters are now committed:
            add_elem_iter_EAR_IDs_cur = copy.deepcopy(self.add_elem_iter_EAR_IDs)
            self._clear_add_elem_iter_EAR_IDs()  # reset to empty nested defaultdict
            for iter_id, all_run_IDs in add_elem_iter_EAR_IDs_cur.items():
                # only re-assign iter_IDs that have not been comitted above:
                if iter_id not in iter_ids:
                    for act_idx, run_IDs in all_run_IDs.items():
                        self.add_elem_iter_EAR_IDs[iter_id][act_idx].extend(run_IDs)

            # pending EARs_initialised that belong to pending iters are now committed:
            self.set_EARs_initialised = [
                i for i in self.set_EARs_initialised if i not in iter_ids
            ]
        self._clear_add_elem_iters()

    @TimeIt.decorator
    @commits_data("add_elem_iter_EAR_IDs")
    def commit_elem_iter_EAR_IDs(self) -> None:
        """
        Commit pending element action run ID updates to disk.
        """
        # TODO: could be batched up?
        for iter_ID, act_EAR_IDs in self.add_elem_iter_EAR_IDs.items():
            self.logger.debug(
                f"commit: adding pending EAR IDs to element iteration {iter_ID!r}: "
                f"{dict(act_EAR_IDs)!r}."
            )
            for act_idx, EAR_IDs in act_EAR_IDs.items():
                self.store._append_elem_iter_EAR_IDs(iter_ID, act_idx, EAR_IDs)
            self.store.element_iter_cache.pop(iter_ID, None)  # invalidate cache
        self._clear_add_elem_iter_EAR_IDs()

    @TimeIt.decorator
    @commits_data("add_EARs")
    def commit_EARs(self) -> None:
        """
        Commit pending element action runs to disk.
        """
        if self.add_EARs:
            EARs = self.store.get_EARs(self.add_EARs)
            EAR_ids = list(self.add_EARs)
            self.logger.debug(f"commit: adding pending EARs with IDs: {EAR_ids!r}")
            self.store._append_EARs(EARs)
            self.store.num_EARs_cache = None  # invalidate cache
            # pending start/end times/snapshots, submission indices, and skips that belong
            # to pending EARs are now committed (accounted for in `get_EARs` above):
            self.set_EAR_submission_data = {
                k: v for k, v in self.set_EAR_submission_data.items() if k not in EAR_ids
            }
            self.set_EAR_skips = {
                k: v for k, v in self.set_EAR_skips.items() if k not in EAR_ids
            }
            self.set_EAR_starts = {
                k: v for k, v in self.set_EAR_starts.items() if k not in EAR_ids
            }
            self.set_EAR_ends = {
                k: v for k, v in self.set_EAR_ends.items() if k not in EAR_ids
            }

        self._clear_add_EARs()

    @TimeIt.decorator
    @commits_data("set_run_dirs")
    def commit_set_run_dirs(self) -> None:
        """
        Commit pending run directory indices.
        """
        for run_dir_arr, run_idx in self.set_run_dirs:
            self.store._set_run_dirs(run_dir_arr, run_idx)
        self._clear_set_run_dirs()

    @TimeIt.decorator
    @commits_data("set_EARs_initialised")
    def commit_EARs_initialised(self) -> None:
        """
        Commit pending element action run init state updates to disk.
        """
        if self.set_EARs_initialised:
            iter_ids = self.set_EARs_initialised
            self.logger.debug(
                f"commit: setting pending `EARs_initialised` for iteration IDs: "
                f"{iter_ids!r}."
            )
            # TODO: could be batched up?
            for i in iter_ids:
                self.store._update_elem_iter_EARs_initialised(i)
                self.store.element_iter_cache.pop(i, None)  # invalidate cache
        self._clear_set_EARs_initialised()

    @TimeIt.decorator
    @commits_data("set_EAR_submission_data")
    def commit_EAR_submission_indices(self) -> None:
        """
        Commit pending element action run submission index updates to disk.
        """
        if self.set_EAR_submission_data:
            self.logger.debug(
                f"commit: updating submission data: {self.set_EAR_submission_data!r}."
            )
            self.store._update_EAR_submission_data(self.set_EAR_submission_data)
            for EAR_ID_i in self.set_EAR_submission_data:
                self.store.EAR_cache.pop(EAR_ID_i, None)  # invalidate cache
            self._clear_EAR_submission_data()

    @TimeIt.decorator
    @commits_data("set_EAR_starts")
    def commit_EAR_starts(self) -> None:
        """
        Commit pending element action run start information to disk.
        """
        updates = self.set_EAR_starts
        if updates:
            self.logger.debug(
                f"commit: registering {len(updates)} run(s) as started: "
                f"{shorten_list_str(updates)}."
            )
            self.store._update_EAR_start(updates)
            for run_id in updates:
                self.store.EAR_cache.pop(run_id, None)  # invalidate cache
        self._clear_set_EAR_starts()

    @TimeIt.decorator
    @commits_data("set_EAR_ends")
    def commit_EAR_ends(self) -> None:
        """
        Commit pending element action run finish information to disk.
        """
        updates = self.set_EAR_ends
        if updates:
            self.logger.debug(
                f"commit: registering {len(updates)} run(s) as ended: "
                f"{shorten_list_str(updates)}, with exit codes: "
                f"{shorten_list_str([i[2] for i in updates.values()])}."
            )
            self.store._update_EAR_end(updates)
            for run_id in updates:
                self.store.EAR_cache.pop(run_id, None)  # invalidate cache
        self._clear_set_EAR_ends()

    @TimeIt.decorator
    @commits_data("set_EAR_skips")
    def commit_EAR_skips(self) -> None:
        """
        Commit pending element action skip flags to disk.
        """
        updates = self.set_EAR_skips
        if updates:
            self.logger.debug(f"commit: setting {len(updates)} run IDs as skipped.")
            self.store._update_EAR_skip(updates)
            for run_ID in updates:
                self.store.EAR_cache.pop(run_ID, None)  # invalidate cache
        self._clear_set_EAR_skips()

    @TimeIt.decorator
    @commits_data("set_js_metadata")
    def commit_js_metadata(self) -> None:
        """
        Commit pending jobscript metadata changes to disk.
        """
        if self.set_js_metadata:
            self.logger.debug(
                f"commit: setting jobscript metadata: {self.set_js_metadata!r}"
            )
            self.store._update_js_metadata(self.set_js_metadata)
        self._clear_set_js_metadata()

    @TimeIt.decorator
    @commits_data("add_parameters", "set_parameters")
    def commit_parameters(self) -> None:
        """Make pending parameters persistent."""
        if self.add_parameters:
            params = self.store.get_parameters(self.add_parameters)
            param_ids = list(self.add_parameters)
            self.logger.debug(f"commit: adding pending parameters IDs: {param_ids!r}")
            self.store._append_parameters(params)
        self._clear_add_parameters()

        if self.set_parameters:
            param_ids = list(self.set_parameters)
            self.logger.debug(f"commit: setting values of parameter IDs {param_ids!r}.")
            self.store._set_parameter_values(self.set_parameters)
            for id_i in param_ids:
                self.store.parameter_cache.pop(id_i, None)
        self._clear_set_parameters()

    @TimeIt.decorator
    @commits_data("add_files")
    def commit_files(self) -> None:
        """Add pending files to the files directory."""
        if self.add_files:
            self.logger.debug("commit: adding pending files to the files directory.")
            self.store._append_files(self.add_files)
        self._clear_add_files()

    @TimeIt.decorator
    @commits_data("add_template_components")
    def commit_template_components(self) -> None:
        """
        Commit pending template components to disk.
        """
        if self.add_template_components:
            self.logger.debug("commit: adding template components.")
            self.store._update_template_components(self.store.get_template_components())
        self._clear_add_template_components()

    @TimeIt.decorator
    @commits_data("update_param_sources")
    def commit_param_sources(self) -> None:
        """Make pending changes to parameter sources persistent."""
        if self.update_param_sources:
            param_ids = list(self.update_param_sources)
            self.logger.debug(f"commit: updating sources of parameter IDs {param_ids!r}.")
            self.store._update_parameter_sources(self.update_param_sources)
            for id_i in param_ids:
                self.store.param_sources_cache.pop(id_i, None)  # invalidate cache
            self._clear_update_param_sources()

    @TimeIt.decorator
    @commits_data("update_loop_indices")
    def commit_loop_indices(self) -> None:
        """Make pending update to element iteration loop indices persistent."""
        updates = self.update_loop_indices
        if updates:
            self.logger.debug(
                f"commit: updating loop indices of {len(updates)} iteration(s)."
            )
            self.store._update_loop_index(updates)
            for iter_ID in updates:
                self.store.element_iter_cache.pop(iter_ID, None)  # invalidate cache
        self._clear_update_loop_indices()

    @TimeIt.decorator
    @commits_data("update_loop_num_iters")
    def commit_loop_num_iters(self) -> None:
        """Make pending update to the number of loop iterations."""
        for index, num_iters in self.update_loop_num_iters.items():
            self.logger.debug(
                f"commit: updating loop {index!r} number of iterations to {num_iters!r}."
            )
            self.store._update_loop_num_iters(index, num_iters)
        self._clear_update_loop_num_iters()

    @TimeIt.decorator
    @commits_data("update_loop_parents")
    def commit_loop_parents(self) -> None:
        """Make pending update to additional loop parents."""
        for index, parents in self.update_loop_parents.items():
            self.logger.debug(f"commit: updating loop {index!r} parents to {parents!r}.")
            self.store._update_loop_parents(index, parents)
        self._clear_update_loop_parents()

    @TimeIt.decorator
    @commits_data("update_iter_data_idx")
    def commit_iter_data_idx(self) -> None:
        if self.update_iter_data_idx:
            self.store._update_iter_data_indices(self.update_iter_data_idx)
            self._clear_update_iter_data_idx()

    @TimeIt.decorator
    @commits_data("update_run_data_idx")
    def commit_run_data_idx(self) -> None:
        if self.update_run_data_idx:
            self.store._update_run_data_indices(self.update_run_data_idx)
            self._clear_update_run_data_idx()

    def _clear_add_tasks(self) -> None:
        self.add_tasks = {}

    def _clear_add_loops(self) -> None:
        self.add_loops = {}

    def _clear_add_submissions(self) -> None:
        self.add_submissions = {}

    def _clear_at_submit_metadata(self) -> None:
        self.update_at_submit_metadata = defaultdict(
            lambda: {i: None for i in SUBMISSION_SUBMIT_TIME_KEYS}
        )

    def _clear_add_elements(self) -> None:
        self.add_elements = {}

    def _clear_add_element_sets(self) -> None:
        self.add_element_sets = defaultdict(list)

    def _clear_add_elem_iters(self) -> None:
        self.add_elem_iters = {}

    def _clear_add_EARs(self) -> None:
        self.add_EARs = {}

    def _clear_set_run_dirs(self):
        self.set_run_dirs = []

    def _clear_add_elem_IDs(self) -> None:
        self.add_elem_IDs = defaultdict(list)

    def _clear_add_elem_iter_IDs(self) -> None:
        self.add_elem_iter_IDs = defaultdict(list)

    def _clear_add_elem_iter_EAR_IDs(self) -> None:
        self.add_elem_iter_EAR_IDs = defaultdict(lambda: defaultdict(list))

    def _clear_set_EARs_initialised(self) -> None:
        self.set_EARs_initialised = []

    def _clear_EAR_submission_data(self) -> None:
        self.set_EAR_submission_data = {}

    def _clear_set_EAR_starts(self) -> None:
        self.set_EAR_starts = {}

    def _clear_set_EAR_ends(self) -> None:
        self.set_EAR_ends = {}

    def _clear_set_EAR_skips(self) -> None:
        self.set_EAR_skips = {}

    def _clear_set_js_metadata(self) -> None:
        self.set_js_metadata = defaultdict(lambda: defaultdict(dict))

    def _clear_add_parameters(self) -> None:
        self.add_parameters = {}

    def _clear_add_files(self) -> None:
        self.add_files = []

    def _clear_add_template_components(self) -> None:
        self.add_template_components = defaultdict(dict)

    def _clear_set_parameters(self) -> None:
        self.set_parameters = {}

    def _clear_update_param_sources(self) -> None:
        self.update_param_sources = {}

    def _clear_update_loop_indices(self) -> None:
        self.update_loop_indices = defaultdict(dict)

    def _clear_update_loop_num_iters(self) -> None:
        self.update_loop_num_iters = {}

    def _clear_update_loop_parents(self) -> None:
        self.update_loop_parents = {}

    def _clear_update_iter_data_idx(self):
        self.update_iter_data_idx = defaultdict(dict)

    def _clear_update_run_data_idx(self):
        self.update_run_data_idx = defaultdict(dict)

    def reset(self, is_init: bool = False) -> None:
        """Clear all pending data and prepare to accept new pending data."""

        if not is_init and not self:
            # no pending changes
            return

        if not is_init:
            self.logger.info("resetting pending changes.")

        self._clear_add_tasks()
        self._clear_add_loops()
        self._clear_add_submissions()
        self._clear_at_submit_metadata()
        self._clear_add_elements()
        self._clear_add_element_sets()
        self._clear_add_elem_iters()
        self._clear_add_EARs()
        self._clear_set_run_dirs()

        self._clear_set_EARs_initialised()
        self._clear_add_elem_IDs()
        self._clear_add_elem_iter_IDs()
        self._clear_add_elem_iter_EAR_IDs()

        self._clear_add_parameters()
        self._clear_add_files()
        self._clear_add_template_components()

        self._clear_EAR_submission_data()
        self._clear_set_EAR_starts()
        self._clear_set_EAR_ends()
        self._clear_set_EAR_skips()

        self._clear_set_js_metadata()
        self._clear_set_parameters()

        self._clear_update_param_sources()
        self._clear_update_loop_indices()
        self._clear_update_loop_num_iters()
        self._clear_update_loop_parents()
        self._clear_update_iter_data_idx()
        self._clear_update_run_data_idx()


@dataclass
class CommitResourceMap:
    """
    Map of :py:class:`PendingChanges` commit method names to store resource labels,
    representing the store resources required by each ``commit_*`` method, for a given
    :py:class:`~.PersistentStore`.

    When :py:meth:`PendingChanges.commit_all` is called, the resources specified will be
    opened in "update" mode, for each ``commit_*`` method.

    Notes
    -----
    Normally only of interest to implementations of persistent stores.
    """

    #: Resources for :py:meth:`~.PendingChanges.commit_tasks`.
    commit_tasks: tuple[str, ...] | None = tuple()
    #: Resources for :py:meth:`~.PendingChanges.commit_loops`.
    commit_loops: tuple[str, ...] | None = tuple()
    #: Resources for :py:meth:`~.PendingChanges.commit_submissions`.
    commit_submissions: tuple[str, ...] | None = tuple()
    #: Resources for :py:meth:`~.PendingChanges.commit_at_submit_metadata`.
    commit_at_submit_metadata: tuple[str, ...] | None = tuple()
    #: Resources for :py:meth:`~.PendingChanges.commit_elem_IDs`.
    commit_elem_IDs: tuple[str, ...] | None = tuple()
    #: Resources for :py:meth:`~.PendingChanges.commit_elements`.
    commit_elements: tuple[str, ...] | None = tuple()
    #: Resources for :py:meth:`~.PendingChanges.commit_element_sets`.
    commit_element_sets: tuple[str, ...] | None = tuple()
    #: Resources for :py:meth:`~.PendingChanges.commit_elem_iter_IDs`.
    commit_elem_iter_IDs: tuple[str, ...] | None = tuple()
    #: Resources for :py:meth:`~.PendingChanges.commit_elem_iters`.
    commit_elem_iters: tuple[str, ...] | None = tuple()
    #: Resources for :py:meth:`~.PendingChanges.commit_elem_iter_EAR_IDs`.
    commit_elem_iter_EAR_IDs: tuple[str, ...] | None = tuple()
    #: Resources for :py:meth:`~.PendingChanges.commit_EARs_initialised`.
    commit_EARs_initialised: tuple[str, ...] | None = tuple()
    #: Resources for :py:meth:`~.PendingChanges.commit_EARs`.
    commit_EARs: tuple[str, ...] | None = tuple()
    #: Resources for :py:meth:`~.PendingChanges.commit_EAR_submission_indices`.
    commit_EAR_submission_indices: tuple[str, ...] | None = tuple()
    #: Resources for :py:meth:`~.PendingChanges.commit_EAR_skips`.
    commit_EAR_skips: tuple[str, ...] | None = tuple()
    #: Resources for :py:meth:`~.PendingChanges.commit_EAR_starts`.
    commit_EAR_starts: tuple[str, ...] | None = tuple()
    #: Resources for :py:meth:`~.PendingChanges.commit_EAR_ends`.
    commit_EAR_ends: tuple[str, ...] | None = tuple()
    #: Resources for :py:meth:`~.PendingChanges.commit_js_metadata`.
    commit_js_metadata: tuple[str, ...] | None = tuple()
    #: Resources for :py:meth:`~.PendingChanges.commit_parameters`.
    commit_parameters: tuple[str, ...] | None = tuple()
    #: Resources for :py:meth:`~.PendingChanges.commit_files`.
    commit_files: tuple[str, ...] | None = tuple()
    #: Resources for :py:meth:`~.PendingChanges.commit_template_components`.
    commit_template_components: tuple[str, ...] | None = tuple()
    #: Resources for :py:meth:`~.PendingChanges.commit_param_sources`.
    commit_param_sources: tuple[str, ...] | None = tuple()
    #: Resources for :py:meth:`~.PendingChanges.commit_loop_indices`.
    commit_loop_indices: tuple[str, ...] | None = tuple()
    #: Resources for :py:meth:`~.PendingChanges.commit_loop_num_iters`.
    commit_loop_num_iters: tuple[str, ...] | None = tuple()
    #: Resources for :py:meth:`~.PendingChanges.commit_loop_parents`.
    commit_loop_parents: tuple[str, ...] | None = tuple()
    #: Resources for :py:meth:`~.PendingChanges.commit_set_run_dirs`.
    commit_set_run_dirs: tuple[str, ...] | None = tuple()
    #: Resources for :py:meth:`~.PendingChanges.commit_iter_data_idx`.
    commit_iter_data_idx: tuple[str, ...] | None = tuple()
    #: Resources for :py:meth:`~.PendingChanges.commit_run_data_idx`.
    commit_run_data_idx: tuple[str, ...] | None = tuple()

    #: A dict whose keys are tuples of resource labels and whose values are lists
    #: of :py:class:`PendingChanges` commit method names that require those resources.
    #:
    #: This grouping allows us to batch up commit methods by resource requirements,
    #: which in turn means we can potentially minimise, e.g., the number of network
    #: requests.
    groups: Mapping[tuple[str, ...], Sequence[str]] = field(
        init=False, repr=False, compare=False
    )

    def __post_init__(self):
        self.groups = self._group_by_resource()

    def _group_by_resource(self) -> Mapping[tuple[str, ...], Sequence[str]]:
        """
        Get a dict whose keys are tuples of resource labels and whose values are
        lists of :py:class:`PendingChanges` commit method names that require those
        resource.

        This grouping allows us to batch up commit methods by resource requirements,
        which in turn means we can potentially minimise e.g. the number of network
        requests.
        """
        groups: dict[tuple[str, ...], list[str]] = {}
        # The dicts are pretending to be insertion-ordered sets
        cur_res_group: tuple[dict[str, None], list[str]] | None = None
        for fld in fields(self):
            if not fld.name.startswith("commit_"):
                continue
            res_labels = getattr(self, fld.name)

            if not cur_res_group:
                # start a new resource group: a mapping between resource labels and the
                # commit methods that require those resources:
                cur_res_group = (dict.fromkeys(res_labels), [fld.name])

            elif not res_labels or set(res_labels).intersection(cur_res_group[0]):
                # there is some overlap between resource labels required in the current
                # group and this commit method, so we merge resource labels and add the
                # new commit method:
                cur_res_group[0].update(dict.fromkeys(res_labels))
                cur_res_group[1].append(fld.name)

            else:
                # no overlap between resource labels required in the current group and
                # those required by this commit method, so append the current group, and
                # start a new group for this commit method:
                groups.setdefault(tuple(cur_res_group[0]), []).extend(cur_res_group[1])
                cur_res_group = (dict.fromkeys(res_labels), [fld.name])

        if cur_res_group:
            groups.setdefault(tuple(cur_res_group[0]), []).extend(cur_res_group[1])

        return groups



import os
import time

import pytest
from hpcflow.app import app as hf


@pytest.mark.integration
@pytest.mark.skipif("hf.run_time_info.is_frozen")
def test_input_file_generator_creates_file(null_config, tmp_path):

    inp_file = hf.FileSpec(label="my_input_file", name="my_input_file.txt")

    if os.name == "nt":
        cmd = "Get-Content <<file:my_input_file>>"
    else:
        cmd = "cat <<file:my_input_file>>"

    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput(parameter=hf.Parameter("p1"))],
        actions=[
            hf.Action(
                commands=[hf.Command(cmd)],
                input_file_generators=[
                    hf.InputFileGenerator(
                        input_file=inp_file,
                        inputs=[hf.Parameter("p1")],
                        script="<<script:input_file_generator_basic.py>>",
                    ),
                ],
                environments=[hf.ActionEnvironment(environment="python_env")],
            )
        ],
    )
    p1_val = 101
    t1 = hf.Task(schema=s1, inputs={"p1": p1_val})
    wk = hf.Workflow.from_template_data(
        tasks=[t1],
        template_name="input_file_generator_test",
        path=tmp_path,
    )
    wk.submit(wait=True, add_to_known=False)

    # check the input file is written
    run_0 = wk.get_all_EARs()[0]
    exec_path = run_0.get_directory()
    inp_file_path = exec_path.joinpath(inp_file.name.name)
    inp_file_contents = inp_file_path.read_text()
    assert inp_file_contents.strip() == str(p1_val)

    # check the command successfully printed the file contents to stdout:
    std_out = wk.submissions[0].jobscripts[0].direct_stdout_path.read_text()
    assert std_out.strip() == str(p1_val)


@pytest.mark.integration
@pytest.mark.skipif("hf.run_time_info.is_frozen")
def test_IFG_std_stream_redirect_on_exception(new_null_config, tmp_path):
    """Test exceptions raised by the app during execution of a IFG script are printed to the
    std-stream redirect file (and not the jobscript's standard error file)."""

    # define a custom python environment which redefines the `WK_PATH` shell variable to
    # a nonsense value so the app cannot load the workflow and thus raises an exception

    app_caps = hf.package_name.upper()
    if os.name == "nt":
        env_cmd = f'$env:{app_caps}_WK_PATH = "nonsense_path"'
    else:
        env_cmd = f'export {app_caps}_WK_PATH="nonsense_path"'

    env_cmd += "; python <<script_path>> <<args>>"
    bad_env = hf.Environment(
        name="bad_python_env",
        executables=[
            hf.Executable(
                label="python_script",
                instances=[
                    hf.ExecutableInstance(
                        command=env_cmd,
                        num_cores=1,
                        parallel_mode=None,
                    )
                ],
            )
        ],
    )
    hf.envs.add_object(bad_env, skip_duplicates=True)

    inp_file = hf.FileSpec(label="my_input_file", name="my_input_file.txt")
    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput(parameter=hf.Parameter("p1"))],
        actions=[
            hf.Action(
                input_file_generators=[
                    hf.InputFileGenerator(
                        input_file=inp_file,
                        inputs=[hf.Parameter("p1")],
                        script="<<script:input_file_generator_basic.py>>",
                    ),
                ],
                environments=[hf.ActionEnvironment(environment="bad_python_env")],
            )
        ],
    )

    p1_val = 101
    t1 = hf.Task(schema=s1, inputs={"p1": p1_val})
    wk = hf.Workflow.from_template_data(
        tasks=[t1], template_name="input_file_generator_test", path=tmp_path
    )
    wk.submit(wait=True, add_to_known=False, status=False)

    # jobscript stderr should be empty
    assert not wk.submissions[0].jobscripts[0].direct_stderr_path.read_text()

    # std stream file has workflow not found traceback
    run = wk.get_all_EARs()[0]
    std_stream_path = run.get_app_std_path()
    assert std_stream_path.is_file()
    assert "WorkflowNotFoundError" in std_stream_path.read_text()

    hf.reload_template_components()  # remove extra envs


@pytest.mark.integration
@pytest.mark.skipif("hf.run_time_info.is_frozen")
def test_IFG_std_out_std_err_not_redirected(null_config, tmp_path):
    """Test that standard error and output streams from an IFG script are written to the jobscript
    standard error and output files."""
    inp_file = hf.FileSpec(label="my_input_file", name="my_input_file.txt")
    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput(parameter=hf.Parameter("p1"))],
        actions=[
            hf.Action(
                input_file_generators=[
                    hf.InputFileGenerator(
                        input_file=inp_file,
                        inputs=[hf.Parameter("p1")],
                        script="<<script:input_file_generator_test_stdout_stderr.py>>",
                    ),
                ],
                environments=[hf.ActionEnvironment(environment="python_env")],
            )
        ],
    )
    p1_val = 101
    stdout_msg = str(p1_val)
    stderr_msg = str(p1_val)
    t1 = hf.Task(schema=s1, inputs={"p1": p1_val})
    wk = hf.Workflow.from_template_data(
        tasks=[t1],
        template_name="input_file_generator_test",
        path=tmp_path,
    )
    wk.submit(wait=True, add_to_known=False)

    if wk.submissions[0].jobscripts[0].resources.combine_jobscript_std:
        std_out_err = wk.submissions[0].jobscripts[0].direct_std_out_err_path.read_text()
        assert std_out_err.strip() == f"{stdout_msg}\n{stderr_msg}"
    else:
        std_out = wk.submissions[0].jobscripts[0].direct_stdout_path.read_text()
        std_err = wk.submissions[0].jobscripts[0].direct_stderr_path.read_text()
        assert std_out.strip() == stdout_msg
        assert std_err.strip() == stderr_msg


@pytest.mark.integration
@pytest.mark.skipif("hf.run_time_info.is_frozen")
def test_IFG_pass_env_spec(null_config, tmp_path):
    inp_file = hf.FileSpec(label="my_input_file", name="my_input_file.txt")

    if os.name == "nt":
        cmd = "Get-Content <<file:my_input_file>>"
    else:
        cmd = "cat <<file:my_input_file>>"

    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput(parameter=hf.Parameter("p1"))],
        actions=[
            hf.Action(
                commands=[hf.Command(cmd)],
                input_file_generators=[
                    hf.InputFileGenerator(
                        input_file=inp_file,
                        inputs=[hf.Parameter("p1")],
                        script="<<script:env_specifier_test/input_file_generator_pass_env_spec.py>>",
                        script_pass_env_spec=True,
                    ),
                ],
                environments=[hf.ActionEnvironment(environment="python_env")],
            )
        ],
    )
    p1_val = 101
    t1 = hf.Task(schema=s1, inputs={"p1": p1_val})
    wk = hf.Workflow.from_template_data(
        tasks=[t1],
        template_name="input_file_generator_pass_env_spec",
        path=tmp_path,
    )
    wk.submit(wait=True, add_to_known=False, status=False)

    # check the command successfully printed the env spec and file contents to stdout:
    std_out = wk.submissions[0].jobscripts[0].direct_stdout_path.read_text()
    assert std_out.strip() == f"{{'name': 'python_env'}}\n{str(p1_val)}"


@pytest.mark.integration
@pytest.mark.skipif("hf.run_time_info.is_frozen")
def test_env_specifier_in_input_file_generator_script_path(new_null_config, tmp_path):

    py_env = hf.Environment(
        name="python_env",
        specifiers={"version": "v1"},
        executables=[
            hf.Executable(
                label="python_script",
                instances=[
                    hf.ExecutableInstance(
                        command="python <<script_path>> <<args>>",
                        num_cores=1,
                        parallel_mode=None,
                    )
                ],
            )
        ],
    )
    hf.envs.add_object(py_env, skip_duplicates=True)

    inp_file = hf.FileSpec(label="my_input_file", name="my_input_file.txt")

    if os.name == "nt":
        cmd = "Get-Content <<file:my_input_file>>"
    else:
        cmd = "cat <<file:my_input_file>>"

    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput(parameter=hf.Parameter("p1"))],
        actions=[
            hf.Action(
                commands=[hf.Command(cmd)],
                input_file_generators=[
                    hf.InputFileGenerator(
                        input_file=inp_file,
                        inputs=[hf.Parameter("p1")],
                        script="<<script:env_specifier_test/<<env:version>>/input_file_generator_basic.py>>",
                    ),
                ],
                environments=[hf.ActionEnvironment(environment="python_env")],
            )
        ],
    )
    p1_val = 101
    t1 = hf.Task(
        schema=s1,
        inputs={"p1": p1_val},
        environments={"python_env": {"version": "v1"}},
    )
    wk = hf.Workflow.from_template_data(
        tasks=[t1],
        template_name="input_file_generator_test_env_specifier",
        path=tmp_path,
    )
    wk.submit(wait=True, add_to_known=False, status=False)

    # check the input file is written
    run_0 = wk.get_all_EARs()[0]
    exec_path = run_0.get_directory()
    inp_file_path = exec_path.joinpath(inp_file.name.name)
    inp_file_contents = inp_file_path.read_text()
    assert inp_file_contents.strip() == str(p1_val)

    # check the command successfully printed the file contents to stdout:
    std_out = wk.submissions[0].jobscripts[0].direct_stdout_path.read_text()
    assert std_out.strip() == str(p1_val)

    hf.reload_template_components()  # remove extra envs
